# -*- coding: utf-8 -*-

import logging

import numpy as np

from ms_peak_picker import (
    FittedPeak, PeakSet, PeakIndex, simple_peak, is_peak)

from ms_deisotope.averagine import (isotopic_shift, neutral_mass)
from ms_deisotope.peak_set import DeconvolutedPeak, Envelope
from ms_deisotope.constants import ERROR_TOLERANCE


logger = logging.getLogger("ms_deisotope.deconvolution")
logger.addHandler(logging.NullHandler())
info = logger.info
debug = logger.debug
error = logger.error


def _is_parallel_arrays(peaks):
    if len(peaks) == 2:
        if isinstance(peaks[0], np.ndarray) and isinstance(peaks[1], np.ndarray) and \
                len(peaks[0]) == len(peaks[1]):
            return True
    return False


def prepare_peaklist(peaks):
    """Ensure ``peaks`` is a :class:`~.PeakSet` object,
    converting from other compatible types as needed. Additionally, make a deep
    copy of the peaks as signal subtraction methods will modify peaks in place.

    This function ensures that any of the following common input types are coerced
    to the appropriate type:

    1. :class:`ms_peak_picker.PeakSet` will be copied and indexed
    2. :class:`ms_peak_picker.PeakIndex` will have its peaks extracted and copied
    3. Any other *sequence* of :class:`PeakLike` objects (objects having an mz and
       intensity attribute) will be converted into a :class:`ms_peak_picker.PeakSet`
    4. A pair of parallel sequences or NumPy arrays corresponding to m/z and intensity
       values for a series of peaks.
    5. Any *sequence* of :class:`tuple` or :class:`list` having at least two entries
       will be converted into a :class:`ms_peak_picker.PeakSet` with the m/z value
       of each peak being the the `p[0]` of each entry and the intensity `p[1]`. Any
       other entries will be ignored.

    Parameters
    ----------
    peaks: Sequence
        Any sequence of :class:`~.FittedPeak` objects, objects
        with ``mz`` and ``intensity`` attributes, or :class:`list` / :class:`tuple`
        objects containing paired values for ``mz`` and ``intensity``

    Returns
    -------
    :class:`~.PeakSet`
    """
    if isinstance(peaks, PeakIndex):
        peaks = PeakSet(peaks.peaks).clone()
    else:
        peaks = tuple(peaks)
        if len(peaks) == 0:
            return PeakSet([])

        elif _is_parallel_arrays(peaks):
            peaks = [simple_peak(x, y) for x, y in zip(*peaks)]

        elif not isinstance(peaks[0], FittedPeak):
            if is_peak(peaks[0]):
                peaks = [simple_peak(p.mz, p.intensity, 0.01) for p in peaks]
            elif isinstance(peaks[0], (list, tuple)):
                peaks = [simple_peak(p[0], p[1], 0.01) for p in peaks]
            else:
                raise TypeError("Cannot convert peaks into a PeakSet")

        peaks = PeakSet(peaks).clone()
    peaks.reindex()
    return peaks


def from_fitted_peak(peak, charge=1):
    """Convert a :class:`~.FittedPeak` into a :class:`~.DeconvolutedPeak`
    at the specified charge state.

    Parameters
    ----------
    peak : :class:`~.FittedPeak`
        The fitted peak to use as the template
    charge : int, optional
        The charge state to use, defaults to 1+

    Returns
    -------
    :class:`~.DeconvolutedPeak`
    """
    mass = neutral_mass(peak.mz, charge)
    dpeak = DeconvolutedPeak(
        mass, peak.intensity, charge,
        peak.signal_to_noise, -1, peak.full_width_at_half_max,
        0, mass, mass, 0, Envelope([(peak.mz, peak.intensity)]),
        peak.mz, area=peak.area)
    return dpeak


def mean(numbers):
    """quick and dirty mean calculation
    without converting to a NumPy array

    Parameters
    ----------
    numbers: Iterable

    Returns
    -------
    float
    """
    n = 0.
    total = 0
    for x in numbers:
        n += 1.
        total += x
    return total / n


def has_previous_peak_at_charge(peak_collection, peak, charge=2, step=1, error_tolerance=ERROR_TOLERANCE):
    """Get the `step`th *preceding* peak from `peak` in a isotopic pattern at
    charge state `charge`, or return `None` if it is missing.

    Parameters
    ----------
    peak_collection : DeconvoluterBase
        Peak collection to look up peaks in. Calls :meth:`has_peak`
    peak : ms_peak_picker.FittedPeak
        The peak to use as a point of reference
    charge : int, optional
        The charge state to interpolate from. Defaults to 2.
    step : int, optional
        The number of peaks along the isotopic pattern to search.

    Returns
    -------
    FittedPeak
    """
    prev = peak.mz - isotopic_shift(charge) * step
    return peak_collection.has_peak(prev, error_tolerance)


def has_successor_peak_at_charge(peak_collection, peak, charge=2, step=1, error_tolerance=ERROR_TOLERANCE):
    """Get the `step`th *succeeding* peak from `peak` in a isotopic pattern at
    charge state `charge`, or return `None` if it is missing.

    Parameters
    ----------
    peak_collection : DeconvoluterBase
        Peak collection to look up peaks in. Calls :meth:`has_peak`
    peak : ms_peak_picker.FittedPeak
        The peak to use as a point of reference
    charge : int, optional
        The charge state to interpolate from. Defaults to 2.
    step : int, optional
        The number of peaks along the isotopic pattern to search.

    Returns
    -------
    FittedPeak
    """
    nxt = peak.mz + isotopic_shift(charge) * step
    return peak_collection.has_peak(nxt, error_tolerance)


def first_peak(peaks):
    """Get the first non-placeholder peak in a list of peaks

    Parameters
    ----------
    peaks : Iterable of FittedPeak

    Returns
    -------
    FittedPeak
    """
    for peak in peaks:
        if peak.intensity > 1 and peak.mz > 1:
            return peak


def drop_placeholders(peaks):
    """Removes all placeholder peaks from an iterable of peaks

    Parameters
    ----------
    peaks : Iterable of FittedPeak

    Returns
    -------
    list
    """
    return [peak for peak in peaks if peak.mz > 1 and peak.intensity > 1]


def count_placeholders(peaks):
    """Counts the number of placeholder peaks in an iterable
    of FittedPeaks

    Parameters
    ----------
    peaks : Iterable of FittedPeak

    Returns
    -------
    int
        Number of placeholder peaks
    """
    i = 0
    for peak in peaks:
        if peak.intensity <= 1:
            i += 1
    return i


def drop_placeholders_parallel(peaks, otherpeaks):
    """Given two parallel iterables of Peak objects, `peaks` and `otherpeaks`,
    for each position that is not a placeholder in `peaks`, include that Peak object
    and its counterpart in `otherpeaks` in a pair of output lists.

    Parameters
    ----------
    peaks : Iterable of FittedPeak
        Peak collection to filter against
    otherpeaks : Iterable
        Collection of objects (Peak-like) to include based upon
        contents of `peaks`

    Returns
    -------
    list
        Filtered form of `peaks`
    list
        Filtered form of `otherpeaks`
    """
    new_peaks = []
    new_otherpeaks = []
    for i, peak in enumerate(peaks):
        peak = peaks[i]
        if peak.intensity > 1:
            new_peaks.append(peak)
            new_otherpeaks.append(otherpeaks[i])
    return new_peaks, new_otherpeaks


def quick_charge(peak_set, index, min_charge, max_charge):
    """An implementation of Hoopman's :title-reference:`QuickCharge` [1] algorithm for quickly capping charge
    state queries

    Parameters
    ----------
    peak_set : :class:`ms_peak_picker.PeakSet
        The centroided peak set to search
    index : int
        The index of the peak to start the search from
    min_charge : int
        The minimum charge state to consider
    max_charge : int
        The maximum charge state to consider

    Returns
    -------
    np.ndarray
        The list of feasible charge states

    References
    ----------
    [1] Hoopmann, M. R., Finney, G. L., MacCoss, M. J., Michael R. Hoopmann, Gregory L. Finney,
        and, MacCoss*, M. J., … MacCoss, M. J. (2007). "High-speed data reduction, feature detection
        and MS/MS spectrum quality assessment of shotgun proteomics data sets using high-resolution
        Mass Spectrometry". Analytical Chemistry, 79(15), 5620–5632. https://doi.org/10.1021/ac0700833
    """
    size = len(peak_set)
    if size == 0:
        raise IndexError("peak_set cannot be empty!")
    if index > size:
        first_index = peak_set[0].peak_count
        if (index - first_index) < size and (index - first_index) >= 0:
            index -= first_index
        else:
            raise IndexError("%d is out of bounds for peak list of size %d in quick_charge" % (index, size))
    min_intensity = peak_set[index].intensity / 4.
    charges = np.zeros(max_charge, dtype=int)
    for j in range(index + 1, size):
        if peak_set[j].intensity < min_intensity:
            continue
        diff = peak_set[j].mz - peak_set[index].mz
        if diff > 1.1:
            break
        raw_charge = 1 / diff
        charge = int(raw_charge + 0.5)
        remain = raw_charge - int(raw_charge)
        if 0.2 < remain < 0.8:
            continue
        if charge < min_charge or charge > max_charge:
            continue
        charges[charge] = 1
    if not np.any(charges):
        return np.array([], dtype=int)
    for j in range(index - 1, -1, -1):
        diff = peak_set[index].mz - peak_set[j].mz
        if diff > 1.1:
            break
        raw_charge = 1 / diff
        charge = int(raw_charge + 0.5)
        remain = raw_charge - int(raw_charge)
        if 0.2 < remain < 0.8:
            continue
        if charge < min_charge or charge > max_charge:
            continue
        charges[charge] = 1
    return np.where(charges)[0]


try:
    _quick_charge = quick_charge
    from ms_deisotope._c.deconvoluter_base import quick_charge
except ImportError:
    pass


def charge_range_(lo, hi, step=None):
    """Generate a sequence of successive charge states, automatically deducing
    the polarity of the charge range to infer the step.

    Parameters
    ----------
    lo : int
        The smallest magnitude charge (positive or negative)
    hi : int
        The largest magnitude charge (positive or negative)
    step : int, optional
        The step size and sign to incrementally alter the charge by. Automatically
        inferred if omitted.

    Yields
    ------
    int
    """
    sign = -1 if lo < 0 else 1
    abs_lo, abs_hi = abs(lo), abs(hi)
    upper = max(abs_lo, abs_hi)
    lower = min(abs_lo, abs_hi)

    for c in range(upper, lower - 1, -1):
        yield c * sign


class ChargeIterator(object):
    def __init__(self, lo, hi):
        self.set_bounds(lo, hi)
        self.make_sequence()

    def set_bounds(self, lo, hi):
        self.sign = -1 if lo < 0 else 1
        abs_lo, abs_hi = abs(lo), abs(hi)
        if abs_lo < abs_hi:
            self.lower = abs_lo
            self.upper = abs_hi
        else:
            self.lower = abs_hi
            self.upper = abs_lo
        self.size = self.upper - self.lower + 1

    def make_sequence(self):
        self.index = 0
        self.size = self.upper - self.lower + 1
        self.values = [self.sign * (self.upper - i) for i in range(self.size)]

    def __len__(self):
        return self.size

    def reset(self):
        self.index = 0

    def sequence_from_quickcharge(self, peak_set, peak):
        charges = quick_charge(peak_set, peak.peak_count,
                               abs(self.lower), abs(self.upper))
        n = charges.shape[0]
        self.index = 0
        if n == 0:
            self.size = 1
            self.values = [1 * self.sign]
        elif charges[0] != 1:
            self.size = n + 1
            self.values = [1 * self.sign] + list(self.sign * charges)
        else:
            self.size = n
            self.values = self.sign * charges

    def __iter__(self):
        return self

    def __next__(self):
        if self.index == self.size:
            raise StopIteration()
        value = self.values[self.index]
        self.index += 1
        return value

    def next(self):
        return self.__next__()



"""A module defining types for describing instrument scans and windows,
as well as a :class:`~.Term` subclass for the `scan attribute` family.
"""

from collections import namedtuple
from collections.abc import MutableSequence
from typing import Optional

from ms_deisotope.utils import _MappingOverAttributeProxy
from .cv import Term, TermSet


def _isclose(x, y, atol=1e-3):
    return abs(x - y) < atol


_IsolationWindowBase = namedtuple(
    "IsolationWindow", ['lower', 'target', 'upper'])


class IsolationWindow(_IsolationWindowBase):
    r"""Describes the m/z interval a precursor ion was isolated from in the precursor scan

    An :class:`IsolationWindow` instance is comparable, hashable, and orderable. It is based
    on a :class:`namedtuple`, and supports the same interface and methods as a :class:`tuple`.

    Attributes
    ----------
    lower: float
        The distance from the center of the isolation window towards 0
    target: float
        The center of the isolation window in m/z
    upper: float
        The distance from the center of the isolation window towards :math:`\infty`
    lower_bound: float
        The m/z coordinate of the lower bound :attr:`target` - :attr:`lower`
    upper_bound: float
        The m/z coordinate of the upper bound :attr:`target` + :attr:`upper`
    width: float
        The sum of :attr:`lower` and :attr:`upper`, the total m/z space spanned by
        the window
    """

    __slots__ = ()
    __hash__ = _IsolationWindowBase.__hash__

    @classmethod
    def make_empty(cls, point):
        """A helper method for instantiating an empty isolation window centered at
        `point`

        Parameters
        ----------
        point : float
            The point to center the new empty window at.

        Returns
        -------
        :class:`IsolationWindow`
        """
        return cls(0, point, 0)

    @property
    def lower_bound(self):
        """The m/z coordinate of the lower bound :attr:`target` - :attr:`lower`
        """
        return self.target - self.lower

    @property
    def upper_bound(self):
        """The m/z coordinate of the upper bound :attr:`target` + :attr:`upper`
        """
        return self.target + self.upper

    @property
    def width(self):
        """The sum of :attr:`lower` and :attr:`upper`, the total m/z space spanned by
        the window"""
        return self.lower + self.upper

    def __contains__(self, x):
        return self.spans(x, 0.1)

    def spans(self, x, tolerance=0.1):
        """Test whether `x` is within the interval defined by this window's bounds.

        This method permits a small amount of error (controlled by `tolerance`) when
        computing the bounds.

        Equivalent to: :code:`(self.lower_bound - tolerance) <= x <= (self.upper_bound + tolerance)`

        Parameters
        ----------
        x : float
            The number to query
        tolerance : float, optional
            The amount of error to accept when computing the bounds (the default is 0.1)

        Returns
        -------
        :class:`bool`
        """
        return (self.lower_bound - tolerance) <= x <= (self.upper_bound + tolerance)

    def is_empty(self):
        """Tests if the window is empty (e.g. its upper bound is equal to its lower bound)

        Returns
        -------
        :class:`bool`
        """
        if self.lower is None:
            return self.upper is None
        return self.lower == self.upper == 0.0

    def __nonzero__(self):
        return not self.is_empty()

    def __bool__(self):
        return self.__nonzero__()

    def __eq__(self, other):
        return _isclose(self.lower, other.lower) and _isclose(
            self.upper, other.upper) and _isclose(
                self.target, other.target)

    def __ne__(self, other):
        return not (self == other)

    @property
    def __dict__(self):
        return _MappingOverAttributeProxy(self)

    def __reduce__(self):
        return self.__class__, (self.lower, self.target, self.upper)


class ScanAcquisitionInformation(MutableSequence):
    """Describes the set distinct scans along the measurable range by
    the instrument that were performed to produce the acquired data

    :class:`ScanAcquisitionInformation` objects implement the :class:`MutableSequence`
    interface, acting as a sequence of :class:`ScanEventInformation` objects.

    Attributes
    ----------
    combination : :class:`str`
        A controlled vocabulary string describing the way in which scans were combined
    scan_list : :class:`list` of :class:`ScanEventInformation`
        The list of scan events performed
    """

    __slots__ = ("combination", "scan_list")

    @property
    def __dict__(self):
        return _MappingOverAttributeProxy(self)

    def __init__(self, combination, scan_list):  # pylint: disable=super-init-not-called
        self.combination = combination
        self.scan_list = scan_list

    def __getitem__(self, i):
        return self.scan_list[i]

    def __setitem__(self, i, value):
        self.scan_list[i] = value

    def __delitem__(self, i):
        del self.scan_list[i]

    def insert(self, index, value):
        """Insert an event at a given position

        Parameters
        ----------
        index : int
            The position to add the new event
        value : :class:`ScanEventInformation`
            The event to insert
        """
        if not isinstance(value, ScanEventInformation):
            raise TypeError(
                f"Expected ScanEventInformation but got {type(value)}"
            )
        self.scan_list.insert(index, value)

    def __len__(self):
        return len(self.scan_list)

    def __repr__(self):
        return f"ScanAcquisitionInformation(combination={self.combination}, scan_list={self.scan_list})"

    def __eq__(self, other):
        return self.combination == other.combination and self.scan_list == other.scan_list

    def __ne__(self, other):
        return not (self == other)

    def __reduce__(self):
        return self.__class__, (self.combination, self.scan_list)

    def copy(self) -> 'ScanAcquisitionInformation':
        """Make a shallow copy of this object"""
        return self.__class__(
            self.combination, [se.copy() for se in self.scan_list])


class _IonMobilityMixin(object):
    __slots__ = ()

    @property
    def _ion_mobility(self):
        return IonMobilityMethods(self)

    def has_ion_mobility(self):
        """Check whether the scan has ion mobility measure.

        Returns
        -------
        bool
        """
        return self._ion_mobility.has_ion_mobility()

    @property
    def drift_time(self):
        """Fetch the drift time quantity of the scan

        Returns
        -------
        value: float or :const:`None`
            The measured drift time, or :const:`None`
        """
        return self._ion_mobility.drift_time()

    @drift_time.setter
    def drift_time(self, value):
        if value is None or value == 0:
            return
        else:
            raise ValueError("Cannot set drift time directly yet.")

    @property
    def ion_mobility_type(self):
        """Fetch the ion mobility type of the scan.

        Returns
        -------
        ims_type: :class:`ScanAttribute` or :const:`None`
            The ion mobility type, or :const:`None`
        """
        return self._ion_mobility.ion_mobility_type()


class ScanEventInformation(_IonMobilityMixin):
    """Describe a single instrument scan, one or more of which compose
    an actual :class:`Scan` object which describes a spectrum.

    Attributes
    ----------
    start_time: :class:`unitfloat`
        The time since the experiment began when the scan started being acquired.
    window_list: :class:`list` of :class:`ScanWindow`
        The intervals over the m/z axis that were sampled from during this scan.
    injection_time: :class:`unitfloat`
        The time spent injecting ions into the analyzer.
    traits: :class:`dict`
        A mapping from :class:`ScanAttribute` to optional values describing this scan event.
    drift_time: :class:`unitfloat`
        The IMS drift time, if present, :const:`None` otherwise. Determined from :attr:`traits`.
    ion_mobility_type: :class:`ScanAttribute`
        The IMS mechanism type, if present, :const:`None` otherwise. Determined from :attr:`traits`.
    """

    __slots__ = ("start_time", "window_list", "injection_time", "traits")

    @property
    def __dict__(self):
        return _MappingOverAttributeProxy(self)

    def __init__(self, start_time, window_list, drift_time=None,
                 injection_time=None, traits=None, **options):
        self.traits = traits or {}
        self.traits.update(options)
        self.start_time = start_time
        self.window_list = window_list or []
        self.drift_time = drift_time
        self.injection_time = injection_time

    @property
    def scan_configuration(self) -> Optional[str]:
        """Retrieve the scan configuration ID

        Returns
        -------
        str, optional
        """
        return self.traits.get('preset scan configuration')

    def copy(self) -> 'ScanEventInformation':
        """Return a shallow copy of this object"""
        return self.__class__(
            self.start_time, self.window_list[:],
            None, self.injection_time, self.traits.copy())

    def __getitem__(self, i):
        return self.window_list[i]

    def __iter__(self):
        return iter(self.window_list)

    def __len__(self):
        return len(self.window_list)

    def __repr__(self):
        if self.has_ion_mobility():
            tail = f", drift_time={self.drift_time}"
        else:
            tail = ''
        if self.injection_time is not None:
            tail += f', injection_time={self.injection_time}'
        if self.traits:
            tail += f", traits={self.traits}"
        form = f"ScanEventInformation(start_time={self.start_time}, window_list={self.window_list}{tail})"
        return form

    def __eq__(self, other):
        eq = _isclose(self.start_time, other.start_time) and (
            self.window_list == other.window_list)
        if not eq:
            return False
        if self.has_ion_mobility() != other.has_ion_mobility():
            return False
        if self.has_ion_mobility() and not _isclose(self.drift_time, other.drift_time):
            return False
        return True

    def __ne__(self, other):
        return not (self == other)

    def total_scan_window(self):
        """Create a :class:`ScanWindow` spanning from the lowest m/z of the lowest window
        to the highest m/z of the highest window.

        Returns
        -------
        :class:`ScanWindow`
        """
        low = float('inf')
        high = 0
        for window in self:
            low = min(low, window.lower)
            high = max(high, window.upper)
        return ScanWindow(low, high)

    def __reduce__(self):
        return self.__class__, (self.start_time, self.window_list, None, self.injection_time, self.traits)


class ScanWindow(namedtuple("ScanWindow", ['lower', 'upper'])):
    """Represent a single contiguous m/z interval that was sampled from during a scan.

    This type supports spanning testing using the :meth:`__contains__` method, and is
    equality comparable and hashable.

    Attributes
    ----------
    lower: float
        The lower bound of the window
    upper: float
        The upper bound of the window
    """
    __slots__ = ()

    @property
    def __dict__(self):
        return _MappingOverAttributeProxy(self)

    def __reduce__(self):
        return self.__class__, (self.lower, self.upper)

    def __contains__(self, i):
        return self.lower <= i <= self.upper

    def is_empty(self):
        """Test whether the window's bounds are both equal to :const:`None`
        or `0`.

        Returns
        -------
        bool
        """
        if self.lower is None:
            return self.upper is None
        return self.lower == self.upper == 0.0

    def __nonzero__(self):
        return not self.is_empty()

    def __bool__(self):
        return self.__nonzero__()

    def __eq__(self, other):
        return _isclose(self.lower, other.lower) and _isclose(
            self.upper, other.upper)

    def __hash__(self):
        return super(ScanWindow, self).__hash__()

    def __ne__(self, other):
        return not (self == other)


class ScanAttribute(Term):
    """Describes a single trait or attribute belonging to a scan,
    such as injection time, filter string, or instrument configuration.
    """
    __slots__ = ()


scan_attributes = []

# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('scan attribute', term_cls_name="ScanAttribute", writer=cog.out)
# ]]]
# CV Version: 4.1.95
scan_attributes = TermSet([
    ScanAttribute('mass resolution', 'MS:1000011',
                  ('Smallest mass difference between two equal magnitude peaks '
                   'so that the valley between them is a specified fraction of '
                   'the peak height.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('scan rate', 'MS:1000015',
                  ('Rate in Th/sec for scanning analyzers.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('scan start time', 'MS:1000016',
                  ('The time that an analyzer started a scan, relative to the '
                   'start of the MS run.'),
                  'scan attribute',
                  ['scan attribute', 'PSM-level attribute', 'object attribute', 'single identification result attribute', 'identification attribute', 'analysis attribute', 'spectrum identification result details']),
    ScanAttribute('zoom scan', 'MS:1000497',
                  ('Special scan mode where data with improved resolution is '
                   'acquired. This is typically achieved by scanning a more '
                   'narrow m/z window or scanning with a lower scan rate.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('dwell time', 'MS:1000502',
                  ('The time spent gathering data across a peak.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('filter string', 'MS:1000512',
                  ('A string unique to Thermo instrument describing instrument '
                   'settings for the scan.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('preset scan configuration', 'MS:1000616',
                  ('A user-defined scan configuration that specifies the '
                   'instrumental settings in which a spectrum is acquired. An '
                   'instrument may cycle through a list of preset scan '
                   'configurations to acquire data. This is a more generic term '
                   'for the Thermo \\"scan event\\", which is defined in the '
                   'Thermo Xcalibur glossary as: a mass spectrometer scan that '
                   'is defined by choosing the necessary scan parameter '
                   'settings. Multiple scan events can be defined for each '
                   'segment of time.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('mass resolving power', 'MS:1000800',
                  ('The observed mass divided by the difference between two '
                   'masses that can be separated: m/dm. The procedure by which '
                   'dm was obtained and the mass at which the measurement was '
                   'made should be reported.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('analyzer scan offset', 'MS:1000803',
                  ('Offset between two analyzers in a constant neutral loss or '
                   'neutral gain scan. The value corresponds to the neutral loss '
                   'or neutral gain value.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('elution time', 'MS:1000826',
                  ('The time of elution from all used chromatographic columns '
                   '(one or more) in the chromatographic separation step, '
                   'relative to the start of the chromatography.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('interchannel delay', 'MS:1000880',
                  ('The duration of intervals between scanning, during which the '
                   'instrument configuration is switched.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('ion injection time', 'MS:1000927',
                  ('The length of time spent filling an ion trapping device.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('first column elution time', 'MS:1002082',
                  ('The time of elution from the first chromatographic column in '
                   'the chromatographic separation step, relative to the start '
                   'of chromatography on the first column.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('second column elution time', 'MS:1002083',
                  ('The time of elution from the second chromatographic column '
                   'in the chromatographic separation step, relative to the '
                   'start of the chromatography on the second column.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('instrument specific scan attribute', 'MS:1002527',
                  ('Instrument specific scan properties that are associated with '
                   'a value.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('ion mobility attribute', 'MS:1002892',
                  ('An attribute describing ion mobility searches.'),
                  'scan attribute',
                  ['scan attribute', 'PSM-level attribute', 'object attribute', 'single identification result attribute', 'identification attribute', 'analysis attribute', 'spectrum identification result details']),
    ScanAttribute('scan number', 'MS:1003057',
                  ('Ordinal number of the scan indicating its order of '
                   'acquisition within a mass spectrometry acquisition run.'),
                  'scan attribute',
                  ['scan attribute', 'object attribute']),
    ScanAttribute('synchronous prefilter selection', 'MS:1002528',
                  ('Synchronous prefilter selection.'),
                  'scan attribute',
                  ['instrument specific scan attribute', 'scan attribute', 'object attribute']),
    ScanAttribute('FAIMS compensation voltage', 'MS:1001581',
                  ('The DC potential applied to the asymmetric waveform in FAIMS '
                   'that compensates for the difference between high and low '
                   'field mobility of an ion.'),
                  'scan attribute',
                  ['ion mobility attribute', 'scan attribute', 'PSM-level attribute', 'object attribute', 'single identification result attribute', 'identification attribute', 'analysis attribute', 'spectrum identification result details']),
    ScanAttribute('ion mobility drift time', 'MS:1002476',
                  ('Drift time of an ion or spectrum of ions as measured in an '
                   'ion mobility mass spectrometer. This time might refer to the '
                   'central value of a bin into which all ions within a narrow '
                   'range of drift time have been aggregated.'),
                  'scan attribute',
                  ['ion selection attribute', 'ion mobility attribute', 'object attribute', 'scan attribute', 'PSM-level attribute', 'single identification result attribute', 'identification attribute', 'analysis attribute', 'spectrum identification result details']),
    ScanAttribute('inverse reduced ion mobility', 'MS:1002815',
                  ('Ion mobility measurement for an ion or spectrum of ions as '
                   'measured in an ion mobility mass spectrometer. This might '
                   'refer to the central value of a bin into which all ions '
                   'within a narrow range of mobilities have been aggregated.'),
                  'scan attribute',
                  ['ion selection attribute', 'ion mobility attribute', 'object attribute', 'scan attribute', 'PSM-level attribute', 'single identification result attribute', 'identification attribute', 'analysis attribute', 'spectrum identification result details']),
])
# [[[end]]]


FAIMS_compensation_voltage = scan_attributes['MS:1001581']
ion_mobility_drift_time = scan_attributes['MS:1002476']
inverse_reduced_ion_mobility = scan_attributes['MS:1002815']
ion_mobility_attribute = scan_attributes['MS:1002892']

ION_MOBILITY_TYPES = {
    FAIMS_compensation_voltage,
    ion_mobility_drift_time,
    inverse_reduced_ion_mobility,
    ion_mobility_attribute,
}


class IonMobilityMethods(object):
    """Determine the ion mobility measure of a scan event.

    This interface attempts to find FAIMS compensation voltage,
    ion mobility drift time, and inverse reduced ion mobility.

    Attributes
    ----------
    scan_event: :class:`ScanEventInformation`
        The scan to interpret.
    """

    def __init__(self, scan_event):
        self.scan_event = scan_event

    def _get_traits(self):
        return self.scan_event.traits

    def get(self):
        """Find the first ion mobility trait of the scan.

        Returns
        -------
        ims_type: :class:`ScanAttribute`
            The ion mobility type.
        value: float
            The drift time value.
        """
        for ims_type in ION_MOBILITY_TYPES:
            result = self._get_traits().get(ims_type.name)
            if result is not None:
                return (ims_type, result)
        return None

    def has_ion_mobility(self):
        """Check whether the scan has ion mobility measure.

        Returns
        -------
        bool
        """
        return self.get() is not None

    def drift_time(self):
        """Fetch the drift time quantity of the scan

        Returns
        -------
        value: float or :const:`None`
            The measured drift time, or :const:`None`
        """
        _ims_type_value = self.get()
        if _ims_type_value is None:
            return None
        _ims_type, value = _ims_type_value
        return value

    def add_ion_mobility(self, ion_mobility_type, drift_time):
        """Set the drift time value for a specific type of ion mobility on
        the scan event.

        Parameters
        ----------
        ion_mobility_type : :class:`unitfloat`
            The type of ion mobility to set the drift time for
        drift_time : :class:`unitfloat`
            The drift time, with appropriate units.
        """
        self._get_traits()[ion_mobility_type] = drift_time

    def remove_ion_mobility_type(self, ion_mobility_type):
        """Remove a specific type of ion mobility from the scan.

        Parameters
        ----------
        ion_mobility_type : :class:`ScanAttribute`
            The type of ion mobility to remove.
        """
        self._get_traits().pop(ion_mobility_type)

    def ion_mobility_type(self):
        """Fetch the ion mobility type of the scan.

        Returns
        -------
        ims_type: :class:`ScanAttribute` or :const:`None`
            The ion mobility type, or :const:`None`
        """
        _ims_type_value = self.get()
        if _ims_type_value is None:
            return None
        ims_type, _value = _ims_type_value
        return ims_type


binary_data_arrays = []

# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('binary data array', writer=cog.out)
# ]]]
# CV Version: 4.1.95
binary_data_arrays = TermSet([
    Term('m/z array', 'MS:1000514',
         ('A data array of m/z values.'),
         'binary data array',
         ['binary data array']),
    Term('intensity array', 'MS:1000515',
         ('A data array of intensity values.'),
         'binary data array',
         ['binary data array']),
    Term('charge array', 'MS:1000516',
         ('A data array of charge values.'),
         'binary data array',
         ['binary data array']),
    Term('signal to noise array', 'MS:1000517',
         ('A data array of signal-to-noise values.'),
         'binary data array',
         ['binary data array']),
    Term('time array', 'MS:1000595',
         ('A data array of relative time offset values from a reference '
          'time.'),
         'binary data array',
         ['binary data array']),
    Term('wavelength array', 'MS:1000617',
         ('A data array of electromagnetic radiation wavelength values.'),
         'binary data array',
         ['binary data array']),
    Term('non-standard data array', 'MS:1000786',
         ('A data array that contains data not covered by any other '
          'term in this group. Please do not use this term, if the '
          'binary data array type might be commonly used - contact the '
          'PSI-MS working group in order to have another CV term added.'),
         'binary data array',
         ['binary data array']),
    Term('flow rate array', 'MS:1000820',
         ('A data array of flow rate measurements.'),
         'binary data array',
         ['binary data array']),
    Term('pressure array', 'MS:1000821',
         ('A data array of pressure measurements.'),
         'binary data array',
         ['binary data array']),
    Term('temperature array', 'MS:1000822',
         ('A data array of temperature measurements.'),
         'binary data array',
         ['binary data array']),
    Term('mean charge array', 'MS:1002478',
         ('Array of mean charge values where the mean charge is '
          'calculated as a weighted mean of the charges of individual '
          'peaks that are aggregated into a processed spectrum.'),
         'binary data array',
         ['binary data array']),
    Term('resolution array', 'MS:1002529',
         ('A data array of resolution values.'),
         'binary data array',
         ['binary data array']),
    Term('baseline array', 'MS:1002530',
         ('A data array of signal baseline values (the signal in the '
          'absence of analytes).'),
         'binary data array',
         ['binary data array']),
    Term('noise array', 'MS:1002742',
         ('A data array of noise values.'),
         'binary data array',
         ['binary data array']),
    Term('sampled noise m/z array', 'MS:1002743',
         ('A data array of parallel, independent m/z values for a '
          'sampling of noise across a spectrum (typically much smaller '
          'than MS:1000514, the m/z array).'),
         'binary data array',
         ['binary data array']),
    Term('sampled noise intensity array', 'MS:1002744',
         ('A data array of intensity values for the amplitude of noise '
          'variation superposed on the baseline (MS:1002745) across a '
          'spectrum (for use with MS:1002743, sampled noise m/z array).'),
         'binary data array',
         ['binary data array']),
    Term('sampled noise baseline array', 'MS:1002745',
         ('A data array of baseline intensity values (the intensity in '
          'the absence of analytes) for a sampling of noise across a '
          'spectrum (for use with MS:1002743, sampled noise m/z array).'),
         'binary data array',
         ['binary data array']),
    Term('ion mobility array', 'MS:1002893',
         ('Abstract array of ion mobility data values. A more specific '
          'child term concept should be specified in data files to make '
          'precise the nature of the data being provided.'),
         'binary data array',
         ['binary data array']),
    Term('mass array', 'MS:1003143',
         ('A data array of mass values.'),
         'binary data array',
         ['binary data array']),
    Term('scanning quadrupole position lower bound m/z array', 'MS:1003157',
         ('Array of m/z values representing the lower bound m/z of the '
          'quadrupole position at each point in the spectrum.'),
         'binary data array',
         ['binary data array']),
    Term('scanning quadrupole position upper bound m/z array', 'MS:1003158',
         ('Array of m/z values representing the upper bound m/z of the '
          'quadrupole position at each point in the spectrum.'),
         'binary data array',
         ['binary data array']),
    Term('mean ion mobility drift time array', 'MS:1002477',
         ('Array of population mean ion mobility values from a drift '
          'time device, reported in seconds (or milliseconds), '
          'corresponding to a spectrum of individual peaks encoded with '
          'an m/z array.'),
         'binary data array',
         ['ion mobility array', 'binary data array']),
    Term('mean ion mobility array', 'MS:1002816',
         ('Array of population mean ion mobility values (K or K0) based '
          'on ion separation in gaseous phase due to different ion '
          'mobilities under an electric field based on ion size, m/z '
          'and shape, corresponding to a spectrum of individual peaks '
          'encoded with an m/z array.'),
         'binary data array',
         ['ion mobility array', 'binary data array']),
    Term('mean inverse reduced ion mobility array', 'MS:1003006',
         ('Array of population mean ion mobility values based on ion '
          'separation in gaseous phase due to different ion mobilities '
          'under an electric field based on ion size, m/z and shape, '
          'normalized for the local conditions and reported in volt- '
          'second per square centimeter, corresponding to a spectrum of '
          'individual peaks encoded with an m/z array.'),
         'binary data array',
         ['ion mobility array', 'binary data array']),
    Term('raw ion mobility array', 'MS:1003007',
         ('Array of raw ion mobility values (K or K0) based on ion '
          'separation in gaseous phase due to different ion mobilities '
          'under an electric field based on ion size, m/z and shape, '
          'corresponding to a spectrum of individual peaks encoded with '
          'an m/z array.'),
         'binary data array',
         ['ion mobility array', 'binary data array']),
    Term('raw inverse reduced ion mobility array', 'MS:1003008',
         ('Array of raw ion mobility values based on ion separation in '
          'gaseous phase due to different ion mobilities under an '
          'electric field based on ion size, m/z and shape, normalized '
          'for the local conditions and reported in volt-second per '
          'square centimeter, corresponding to a spectrum of individual '
          'peaks encoded with an m/z array.'),
         'binary data array',
         ['ion mobility array', 'binary data array']),
    Term('raw ion mobility drift time array', 'MS:1003153',
         ('Array of raw ion mobility values from a drift time device, '
          'reported in seconds (or milliseconds), corresponding to a '
          'spectrum of individual peaks encoded with an m/z array.'),
         'binary data array',
         ['ion mobility array', 'binary data array']),
    Term('deconvoluted ion mobility array', 'MS:1003154',
         ('Array of ion mobility values (K or K0) based on ion '
          'separation in gaseous phase due to different ion mobilities '
          'under an electric field based on ion size, m/z and shape, as '
          'an average property of an analyte post peak-detection, '
          'weighted charge state reduction, and/or adduct aggregation, '
          'corresponding to a spectrum of individual peaks encoded with '
          'an m/z array.'),
         'binary data array',
         ['ion mobility array', 'binary data array']),
    Term('deconvoluted inverse reduced ion mobility array', 'MS:1003155',
         ('Array of ion mobility values based on ion separation in '
          'gaseous phase due to different ion mobilities under an '
          'electric field based on ion size, m/z and shape, normalized '
          'for the local conditions and reported in volt-second per '
          'square centimeter, as an average property of an analyte post '
          'peak-detection, weighted charge state reduction, and/or '
          'adduct aggregation, corresponding to a spectrum of '
          'individual peaks encoded with an m/z array.'),
         'binary data array',
         ['ion mobility array', 'binary data array']),
    Term('deconvoluted ion mobility drift time array', 'MS:1003156',
         ('Array of mean ion mobility values from a drift time device, '
          'reported in seconds (or milliseconds), as an average '
          'property of an analyte post peak-detection, weighted charge '
          'state reduction, and/or adduct aggregation, corresponding to '
          'a spectrum of individual peaks encoded with an m/z array.'),
         'binary data array',
         ['ion mobility array', 'binary data array']),
])
# [[[end]]]


__all__ = [
    "IsolationWindow", "ScanAcquisitionInformation", "ScanEventInformation",
    "ScanWindow", "ScanAttribute", "scan_attributes",
]



# pragma: no cover
import os
import io

from typing import Iterator, List, Optional, Tuple, Union, Hashable, Deque
import warnings

import numpy as np

from six import string_types as basestring
from ms_deisotope.data_source import Scan, ProcessedScan, ProcessedRandomAccessScanSource
from ms_deisotope.peak_set import DeconvolutedPeak

from ms_deisotope.utils import Base
from ms_deisotope.data_source.common import (
    _SingleScanIteratorImpl, _InterleavedGroupedScanIteratorImpl)
from ms_deisotope.data_source._compression import get_opener
from ms_deisotope.data_source.common import ScanBunch, ScanIterator, ChargeNotProvided, PrecursorInformation
from ms_deisotope.feature_map.scan_index import ExtendedScanIndex


class ScanSerializerBase(object):
    """A common base class for all types which serialize :class:`~.Scan`-like objects to
    disk.

    These types support the context manager protocol, controlling their closing behavior.
    """

    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def save_scan_bunch(self, bunch: ScanBunch, **kwargs):
        """Save a single :class:`~.ScanBunch` to the
        writing stream, recording any information relating
        the precursors and products to each other.

        May call :meth:`save_scan` multiple times.

        Parameters
        ----------
        bunch : :class:`~.ScanBunch`
            The scan bunch to save
        """
        self.save_scan(bunch.precursor, **kwargs)
        for prod in bunch.products:
            self.save_scan(prod, **kwargs)

    def save_scan(self, scan: Scan, **kwargs):
        """Save a single scan to the writing stream.

        Parameters
        ----------
        scan : :class:`ScanBase`
            The scan to save
        """
        raise NotImplementedError()

    def save(self, bunch: Union[Scan, ScanBunch], **kwargs):
        """Save any scan information in `bunch`.

        This method can handle :class:`~.ScanBunch` or :class:`ScanBase`
        instances, dispatching to the appropriate logic.

        Parameters
        ----------
        bunch : :class:`~.ScanBunch` or :class:`ScanBase`
            The scan data to save. May be a collection of related scans or a single scan.

        See Also
        --------
        :meth:`save_scan`
        :meth:`save_scan_bunch`
        """
        if isinstance(bunch, ScanBunch):
            self.save_scan_bunch(bunch, **kwargs)
        else:
            self.save_scan(bunch, **kwargs)

    def close(self):
        """Finish writing scan data, write any pending metadata
        and close the file stream.

        May call :meth:`complete`.
        """
        self.complete()

    def complete(self):
        """Perform any final operations after all spectra have been
        written, adding any trailing content necessary for the format.

        May be called from :meth:`close`, but before the writing stream
        has actually been closed.
        """


class ScanDeserializerBase(object):
    def __init__(self, *args, **kwargs):
        super(ScanDeserializerBase, self).__init__(*args, **kwargs)

    def __next__(self):
        return self.next() # pylint: disable=not-callable

    def next(self):
        raise NotImplementedError()

    def get_scan_by_id(self, scan_id):
        raise NotImplementedError()

    def get_scan_by_time(self, rt, require_ms1=False):
        raise NotImplementedError()

    def get_scan_by_index(self, index):
        raise NotImplementedError()


ScanIterator.register(ScanDeserializerBase)


class SampleRun(Base):

    def __init__(self, name, uuid, completed=True, sample_type=None, **kwargs):
        self.name = name
        self.uuid = uuid
        self.sample_type = sample_type
        self.completed = completed
        self.parameters = kwargs


class LCMSMSQueryInterfaceMixin(object):
    """A mixin class for querying an extended data index on a processed data file."""

    def require_extended_index(self) -> Optional[ExtendedScanIndex]:
        """Require the extended index is initialized.

        Either load the index if it exists, or build a new one.

        The new index is built within a the calling process, and may
        be considerably slower than the multiprocess index building
        tools.


        Returns
        -------
        ExtendedIndex
        """
        if not self.has_extended_index():
            try:
                if self.has_index_file():
                    self.read_index_file()
                else:
                    self.build_extended_index()
            except IOError:
                pass
            except ValueError:
                pass
            self._build_scan_id_to_rt_cache()
        return self.extended_index

    def has_extended_index(self) -> bool:
        """Check if the extended index is initialized."""
        return self.extended_index is not None

    def read_index_file(self, index_path=None):
        """Attempt to read the extended index from the file system.

        This updates :attr:`extended_index`.
        """
        if index_path is None:
            index_path = self._index_file_name
        with get_opener(index_path) as handle:
            if 'b' in handle.mode:
                handle = io.TextIOWrapper(handle, 'utf8')
            self.extended_index = ExtendedScanIndex.deserialize(handle)

    def has_index_file(self) -> bool:
        """Checks if an extended index file exists for this reader.

        Returns
        -------
        bool
        """
        try:
            return os.path.exists(self._index_file_name)
        except (TypeError, AttributeError):
            return False

    @property
    def _index_file_name(self) -> Optional[ExtendedScanIndex]:
        if isinstance(self.source_file, basestring):
            return ExtendedScanIndex.index_file_name(self.source_file)
        else:
            try:
                return ExtendedScanIndex.index_file_name(self.source_file.name)
            except AttributeError:
                return None

    def build_extended_index(self, header_only=True):
        """Build the extended index for this reader.

        This method builds the index in-process and may be slower
        than the multiprocessing builders.
        """
        self.reset()
        indexer = ExtendedScanIndex()
        iterator = self
        if header_only:
            iterator = self.iter_scan_headers()
        if self._has_ms1_scans():
            for bunch in iterator:
                indexer.add_scan_bunch(bunch)
        else:
            for scan in iterator:
                indexer.add_scan(scan)

        self.reset()
        self.extended_index = indexer
        try:
            with open(self._index_file_name, 'w') as handle:
                indexer.serialize(handle)
        except (IOError, OSError, AttributeError, TypeError) as err:
            warnings.warn(str(err))

    def get_index_information_by_scan_id(self, scan_id: str) -> dict:
        """Get the scan description from the extended index."""
        try:
            try:
                return self.extended_index.msn_ids[scan_id]
            except KeyError:
                return self.extended_index.ms1_ids[scan_id]
        except Exception:
            return {}

    def convert_scan_id_to_retention_time(self, scan_id: str) -> float:
        try:
            time = self._scan_id_to_rt[scan_id]
            return time
        except KeyError:
            header = self.get_scan_header_by_id(scan_id)
            return header.scan_time

    def _build_scan_id_to_rt_cache(self):
        if self.has_extended_index():
            for key in self.extended_index.ms1_ids:
                self._scan_id_to_rt[key] = self.extended_index.ms1_ids[
                    key]['scan_time']
            for key in self.extended_index.msn_ids:
                self._scan_id_to_rt[key] = self.extended_index.msn_ids[
                    key]['scan_time']

    def precursor_information(self) -> List[PrecursorInformation]:
        out = []
        for _, info in self.extended_index.msn_ids.items():
            mz = info['mz']
            prec_neutral_mass = info['neutral_mass']
            charge = info['charge']
            if charge == 'ChargeNotProvided':
                charge = ChargeNotProvided
            intensity = info['intensity']
            precursor_scan_id = info['precursor_scan_id']
            product_scan_id = info['product_scan_id']
            orphan = info.get('orphan', False)
            coisolation = info.get('coisolation', [])[:]
            defaulted = info.get('defaulted', False)
            pinfo = PrecursorInformation(
                mz, intensity, charge, precursor_scan_id,
                self, prec_neutral_mass, charge, intensity,
                product_scan_id=product_scan_id, orphan=orphan,
                coisolation=coisolation,
                defaulted=defaulted)
            if prec_neutral_mass is None or charge is None:
                continue
            out.append(pinfo)
        return out

    def ms1_peaks_above(self, mass_threshold: float = 500, intensity_threshold: float = 1000.) -> Deque[Tuple[str, DeconvolutedPeak, Hashable]]:
        """Read out all MS1 peaks from this spectrum source, if they satisfy the requested minimum values.

        Loop over the MS1 index, reading out peaks from the processed peak set whose
        mass exceeds ``mass_threshold`` and whose intensity exceeds ``intensity_threshold``.

        .. note::
            The final element of returned tuples is not guaranteed to be an integer, but only
            some hashable type.

        Returns
        -------
        peak_records : Deque[Tuple[:class:`str`, :class:`~.DeconvolutedPeak`, :class:`Hashable`]]
            A list of peaks associated with the MS1 scan ID they came from and an
            arbitrary hashable value to identify the pair.
        """
        accumulate = Deque()
        for ms1_id in self.extended_index.ms1_ids:
            scan = self.get_scan_by_id(ms1_id)
            for peak in scan.deconvoluted_peak_set:
                if peak.intensity < intensity_threshold or peak.neutral_mass < mass_threshold:
                    continue
                accumulate.append((ms1_id, peak, id(peak)))
        return accumulate

    def ms1_scan_times(self) -> np.ndarray:
        times = sorted(
            [bundle['scan_time'] for bundle in
             self.extended_index.ms1_ids.values()])
        return np.array(times)

    def extract_total_ion_current_chromatogram(self) -> np.ndarray:
        current = []
        for scan_id in self.extended_index.ms1_ids:
            header = self.get_scan_header_by_id(scan_id)
            current.append(header.arrays[1].sum())
        return np.array(current)

    def msms_for(self, query_mass, mass_error_tolerance=1e-5, start_time=None, end_time=None) -> List[PrecursorInformation]:
        out = []
        pinfos = self.extended_index.find_msms_by_precursor_mass(
            query_mass, mass_error_tolerance, bind=self)
        for pinfo in pinfos:
            valid = True
            product = None
            if start_time is not None or end_time is not None:
                product = self.get_scan_header_by_id(pinfo.product_scan_id)
            if start_time is not None and product.scan_time < start_time:
                valid = False
            elif end_time is not None and product.scan_time > end_time:
                valid = False
            if valid:
                out.append(pinfo)
        return out

    def _dispose(self):
        self._scan_id_to_rt.clear()
        self.extended_index.clear()
        return super(LCMSMSQueryInterfaceMixin, self)._dispose()

    def iter_scan_headers(self, iterator=None, grouped=True) -> Iterator[Scan]:
        try:
            if not self._has_ms1_scans():
                grouped = False
        except Exception:
            pass
        self.reset()
        if iterator is None:
            iterator = iter(self._source)

        _make_scan = super(LCMSMSQueryInterfaceMixin, self)._make_scan
        _validate = super(LCMSMSQueryInterfaceMixin, self)._validate

        if grouped:
            impl = _InterleavedGroupedScanIteratorImpl(
                iterator, _make_scan, _validate)
        else:
            impl = _SingleScanIteratorImpl(iterator, _make_scan, _validate)

        for x in impl:
            yield x

        self.reset()

    def get_scan_header_by_id(self, scan_id: str) -> Scan:
        """Retrieve the scan object for the specified scan id. If the
        scan object is still bound and in memory somewhere, a reference
        to that same object will be returned. Otherwise, a new object will
        be created.

        Parameters
        ----------
        scan_id : str
            The unique scan id value to be retrieved

        Returns
        -------
        Scan
        """
        try:
            packed = super(LCMSMSQueryInterfaceMixin, self)._make_scan(
                self._source.get_by_id(scan_id))
            return packed
        except AttributeError as ae:
            raise AttributeError("Could not read attribute (%s) while looking up scan %s" % (
                ae, scan_id))



""" Waters
    MassLynx Python SDK
"""
import sys
import copy
import ctypes
from ctypes import*
from array import *
from .MassLynxRawReader import *
from .MassLynxRawDefs import MassLynxBaseType

import numpy as np

if sys.version_info.major >= 3:
    buf_from_mem = ctypes.pythonapi.PyMemoryView_FromMemory
    buf_from_mem.restype = ctypes.py_object
    buf_from_mem.argtypes = (ctypes.c_void_p, ctypes.c_int, ctypes.c_int)
else:
    buf_from_mem = ctypes.pythonapi.PyBuffer_FromMemory
    buf_from_mem.restype = ctypes.py_object


def make_nd_array(c_pointer, shape, dtype=np.float64, order='C', own_data=True):
    arr_size = np.prod(shape[:]) * np.dtype(dtype).itemsize
    if sys.version_info.major >= 3:
        buffer = buf_from_mem(c_pointer, arr_size, 0x100)
    else:
        buffer = buf_from_mem(c_pointer, arr_size)
    arr_size = np.prod(shape[:]) * np.dtype(dtype).itemsize
    arr = np.ndarray(tuple(shape[:]), dtype, buffer, order=order)
    if own_data and not arr.flags.owndata:
        return arr.copy()
    else:
        return arr


class MassLynxRawScanReader(MassLynxRawReader):
    """Read masslynx scan data"""

    def __init__(self, source ):
        super(MassLynxRawScanReader, self).__init__(source, MassLynxBaseType.SCAN)

    #@classmethod
    #def CreateFromPath( cls, path ):                      # alternative constructor - pass class to constructor
    #    return cls(MassLynxRawReader.fromPath( path, 1 ))                     # initalise with reader

    #@classmethod
    #def CreateFromReader( cls, sourceReader ):                  # alternative constructor - pass class to constructor
    #     return cls(MassLynxRawReader.fromReader( sourceReader, 1 ))                     # initalise with reader

    def ReadScan( self, whichFunction, whichScan ):
        masses = []
        intensities = []

        # create the retrun values
        size = c_int(0)
        pMasses = c_void_p()
        pIntensities = c_void_p()

        # read scan
        readScan = MassLynxRawReader.massLynxDll.readScan
        readScan.argtypes = [c_void_p, c_int, c_int, POINTER(c_void_p), POINTER(c_void_p), POINTER(c_int)]
        super(MassLynxRawScanReader, self).CheckReturnCode( readScan(self._getReader(),whichFunction,whichScan,pMasses,pIntensities,size) )

        # fill the array
        pM = cast(pMasses,POINTER(c_float))
        pI = cast(pIntensities,POINTER(c_float))

        masses = pM[0:size.value]
        intensities = pI[0:size.value]

        # dealocate memory
        #MassLynxRawReader.ReleaseMemory( pMasses)
        #MassLynxRawReader.ReleaseMemory( pIntensities)

        # masses = make_nd_array(pMasses, (size, ), np.float32)
        # intensities = make_nd_array(pIntensities, (size, ), np.float32)
        return masses, intensities

    def ReadScanFlags( self, whichFunction, whichScan ):
        masses = []
        intensities = []
        flags = []

        # create the retrun values
        size = c_int(0)
        pMasses = c_void_p()
        pIntensities = c_void_p()
        pFlags = c_void_p()

        # read scan
        readScanFlags = MassLynxRawReader.massLynxDll.readScanFlags
        readScanFlags.argtypes = [c_void_p, c_int, c_int, POINTER(c_void_p), POINTER(c_void_p), POINTER(c_void_p), POINTER(c_int)]
        super(MassLynxRawScanReader, self).CheckReturnCode( readScanFlags(self._getReader(),whichFunction,whichScan,pMasses,pIntensities,pFlags,size) )

        # fill the array
        pM = cast(pMasses,POINTER(c_float))
        pI = cast(pIntensities,POINTER(c_float))

        masses = pM[0:size.value]
        intensities = pI[0:size.value]

        # check for flags
        if None != pFlags.value:
            pF = cast(pFlags,POINTER(c_byte))
            flags = pF[0:size.value]


        return masses, intensities, flags

    def ReadDriftScan( self, whichFunction, whichScan, whichDrift ):
        masses = []
        intensities = []

        # create the retrun values
        size = c_int(0)
        pMasses = c_void_p()
        pIntensities = c_void_p()
        # read scan
        readDriftScan = MassLynxRawReader.massLynxDll.readDriftScan
        readDriftScan.argtypes = [c_void_p, c_int, c_int, c_int, POINTER(c_void_p), POINTER(c_void_p), POINTER(c_int)]
        try:
            super(MassLynxRawScanReader, self).CheckReturnCode( readDriftScan(self._getReader(),whichFunction, whichScan, whichDrift, pMasses,pIntensities,size) )# fill the array
            pM = cast(pMasses,POINTER(c_float))
            pI = cast(pIntensities,POINTER(c_float))
            masses = pM[0:size.value]
            intensities = pI[0:size.value]

            # masses = make_nd_array(pMasses, (size.value, ), np.float32)
            # intensities = make_nd_array(pIntensities, (size.value, ), np.float32)

            # dealocate memory - causes segfaults?
            # MassLynxRawReader.ReleaseMemory( pMasses)
            # MassLynxRawReader.ReleaseMemory( pIntensities)

            return masses, intensities

        except Exception as err:
            return np.array([], dtype=np.float32), np.array([], dtype=np.float32)

    def ReadDriftScanFlagsIndex(self, whichFunction, whichScan, whichDrift):
        masses = []
        intensities = []
        flags = []

        # create the retrun values
        size = c_int(0)
        pMasses = c_void_p()
        pIntensities = c_void_p()
        pFlags = c_void_p()

        readDriftScanFlagsIndex = MassLynxRawReader.massLynxDll.readDriftScanFlagsIndex
        readDriftScanFlagsIndex.argtypes = [c_void_p, c_int, c_int, c_int, POINTER(c_void_p), POINTER(c_void_p),
                                            POINTER(c_void_p), POINTER(c_int)]

        super(MassLynxRawScanReader, self).CheckReturnCode(
            readDriftScanFlagsIndex(self._getReader(),whichFunction,whichScan,whichDrift,pMasses,pIntensities,pFlags,size))        # fill the array

        pM = cast(pMasses,POINTER(c_float))
        pI = cast(pIntensities,POINTER(c_float))

        masses = pM[0:size.value]
        intensities = pI[0:size.value]

        # check for flags
        if pFlags.value is not None:
            pF = cast(pFlags,POINTER(c_byte))
            flags = pF[0:size.value]
        # Need to manage memory here?
        return masses, intensities, flags


    #def readDaughterScan( self, whichFunction, whichScan ):
    #    try:
    #        size = self.getScanSize( whichFunction,whichScan )

    #       # get the daughter scan size
    #        daughtersize = c_int(0)

    #        # get daughter size
    #        getDaughterScanSize =  RawReader.massLynxDll.getDaughterScanSize
    #        getDaughterScanSize.argtypes = [c_void_p, c_int, c_int, POINTER(c_int)]
    #        RawReader.CheckReturnCode( getDaughterScanSize(RawReader.getReader(self),whichFunction,whichScan, daughtersize) )

       #      # create the float arrays
    #        masses = (c_float*size)()
    #        intensities = (c_float*size)()
    #        daughters = (c_float*daughtersize.value)()

    #        # read daughter size
    #        readSpectrumDaughters = RawReader.massLynxDll.readSpectrumDaughters
    #        readSpectrumDaughters.argtypes = [c_void_p, c_int, c_int,  POINTER(c_float), POINTER(c_float), POINTER(c_float)]
    #        RawReader.CheckReturnCode( readSpectrumDaughters(RawReader.getReader(self), whichFunction, whichScan, masses, intensities, daughters) )

    #    except RawReaderException as e:
    #        e.Handler()
    #        return [], [], []

    #    return list(masses), list(intensities), list(daughters)



"""A module for signature ion extraction.
"""

from collections import namedtuple
from numbers import Number

from dataclasses import dataclass, field
from typing import List

from ms_deisotope.utils import Base
from ms_deisotope.averagine import mass_charge_ratio
from ms_deisotope.data_source.scan.base import BasePeakMethods, PeakSetMethods

class Target(Base):
    """An analyte with known mass and expected charge states

    Attributes
    ----------
    name: str
        The name of the analyte
    neutral_mass: float
        The known neutral mass of the analyte
    charge_states: list
        The list of expected charge states, defaults to ``[1]``

    """
    def __init__(self, name, neutral_mass, charge_states=None):
        if isinstance(charge_states, Number):
            charge_states = [charge_states]
        elif charge_states is None:
            charge_states = [1]
        self.name = name
        self.neutral_mass = neutral_mass
        self.charge_states = charge_states

    def iter_mz(self):
        """Iterate over theoretical m/z coordinates for this analyte

        Yields
        ------
        float:
            The theoretical m/z for a charge state
        """
        for charge in self.charge_states:
            yield mass_charge_ratio(self.neutral_mass, charge)

    def __eq__(self, other):
        return self.name == other.name and self.charge_states == other.charge_states and abs(
            self.neutral_mass - other.neutral_mass) < 1e-3

    def __hash__(self):
        return hash(self.name)


class SignatureIonDetector(object):
    """Detect when a scan contains intense signal for signature ions.

    Implements the :class:`~.Callable` interface.

    Attributes
    ----------
    signature_ions: :class:`list` of :class:`Target`
        The target ions to scan for.
    method: str
        The detection method name, as a string. Currently supports "ratio" only.

    """

    METHODS = ('ratio', )

    signature_ions: List[Target]

    def __init__(self, signature_ions: List[Target]=None, method='ratio'):
        if signature_ions is None:
            signature_ions = []
        self.signature_ions = signature_ions
        if method not in self.METHODS:
            raise ValueError("Method not recognized: %r" % (method, ))
        self.method = method

    def scan(self, peak_list, error_tolerance=2e-5):
        """Scan the peak list for signature ion masses or m/zs.

        Parameters
        ----------
        peak_list : peak set-like
            The peak set to search. May be any object compatible with
            :class:`~.PeakSetMethods`.
        error_tolerance : float
            The PPM error tolerance for peak mass coordinate matching.

        Returns
        -------
        list
        """
        result = []
        peak_list = PeakSetMethods(peak_list)
        if peak_list.is_deconvoluted:
            for sig in self.signature_ions:
                matches = peak_list.all_peaks_for(
                    sig.neutral_mass, error_tolerance)
                for match in matches:
                    if match.charge in sig.charge_states:
                        result.append((sig, match))
        else:
            for sig in self.signature_ions:
                for mz in sig.iter_mz():
                    matches = peak_list.all_peaks_for(mz, error_tolerance)
                    for match in matches:
                        result.append((sig, match))
        return matches

    def ratio(self, peak_list, error_tolerance=2e-5):
        """Compute an intensity ratio of the signature ions against the
        base peak.

        Parameters
        ----------
        peak_list : peak set-like
            The peak set to search. May be any object compatible with
            :class:`~.PeakSetMethods`.
        error_tolerance : float
            The PPM error tolerance for peak mass coordinate matching.

        Returns
        -------
        float
        """
        try:
            base_peak = BasePeakMethods(peak_list)()
            maximum = base_peak.intensity
        except (ValueError, TypeError):
            if len(peak_list) == 0:
                return 0
            else:
                raise
        n = len(self.signature_ions)
        if n == 0:
            return 0
        oxonium = sum(
            p.intensity / maximum for p in self.scan(
                peak_list, error_tolerance))
        return oxonium / n

    def detect(self, peak_list, error_tolerance=2e-5):
        """Test the peak list for the signature ions.

        Parameters
        ----------
        peak_list : peak set-like
            The peak set to search. May be any object compatible with
            :class:`~.PeakSetMethods`.
        error_tolerance : float
            The PPM error tolerance for peak mass coordinate matching.

        Returns
        -------
        float
        """
        if self.method == 'ratio':
            return self.ratio(peak_list, error_tolerance)
        else:
            raise NotImplementedError(self.method)

    def __call__(self, peak_list, error_tolerance=2e-5):
        return self.detect(peak_list, error_tolerance)


@dataclass(frozen=True)
class ReporterIonInfo:
    name: str
    intact_mass: float
    reporter_mass_hcd: float
    reporter_mass_etd: float = field(default=0.0)

TMTInfo = ReporterIonInfo


TMT0_INFO = [
    ReporterIonInfo('TMT0-126', 224.152478, 125.12044953323, 113.12044853323)
]


TMT2_INFO = [
    ReporterIonInfo('TMT2-126', 225.155833, 125.12044953323, 113.12044853323),
    ReporterIonInfo('TMT2-127', 225.155833, 126.12380453323, 113.12044853323),
]


TMT6_INFO = [
    ReporterIonInfo('TMT6-126', 229.162932, 125.12044953323, 113.12044853323),
    ReporterIonInfo('TMT6-127', 229.162932, 126.11748453323001, 114.11748353323),
    ReporterIonInfo('TMT6-128', 229.162932, 127.12715953323, 115.12715653323),
    ReporterIonInfo('TMT6-129', 229.162932, 128.12419453323, 116.12419153323),
    ReporterIonInfo('TMT6-130', 229.162932, 129.13386853323, 117.13386453323001),
    ReporterIonInfo('TMT6-131', 229.162932, 130.13090353323, 118.13089953323001),
]


TMT10_INFO = [
    ReporterIonInfo('TMT10-126', 229.162932, 125.12044953323, 113.12044853323),
    ReporterIonInfo('TMT10-127N', 229.162932, 126.11748453323001, 114.11748353323),
    ReporterIonInfo('TMT10-127C', 229.162932, 126.12380453323, 113.12044853323),
    ReporterIonInfo('TMT10-128N', 229.162932, 127.12083953323001, 114.12044853323),
    ReporterIonInfo('TMT10-128C', 229.162932, 127.12715953323, 115.12715653323),
    ReporterIonInfo('TMT10-129N', 229.162932, 128.12419453323, 116.12419153323),
    ReporterIonInfo('TMT10-129C', 229.162932, 128.13051353323, 115.12715653323),
    ReporterIonInfo('TMT10-130N', 229.162932, 129.12754853323, 116.12419153323),
    ReporterIonInfo('TMT10-130C', 229.162932, 129.13386853323, 117.13386453323001),
    ReporterIonInfo('TMT10-131', 229.162932, 130.13090353323, 118.13089953323001),
]


TMT11_INFO = TMT10_INFO[:]
TMT11_INFO.append(
    ReporterIonInfo('TMT11-131C', 229.162932, 130.13722253323, 117.13386453323001))


TMT16_INFO = [
    ReporterIonInfo('TMTpro-126', 304.207100, 126.127726, 0.0),
    ReporterIonInfo('TMTpro-127N', 304.207100, 127.124761, 0.0),
    ReporterIonInfo('TMTpro-127C', 304.207100, 127.131081, 0.0),
    ReporterIonInfo('TMTpro-128N', 304.207100, 128.128116, 0.0),
    ReporterIonInfo('TMTpro-128C', 304.207100, 128.134436, 0.0),
    ReporterIonInfo('TMTpro-129N', 304.207100, 129.131471, 0.0),
    ReporterIonInfo('TMTpro-129C', 304.207100, 129.137790, 0.0),
    ReporterIonInfo('TMTpro-130N', 304.207100, 130.134825, 0.0),
    ReporterIonInfo('TMTpro-130C', 304.207100, 130.141145, 0.0),
    ReporterIonInfo('TMTpro-131N', 304.207100, 131.138180, 0.0),
    ReporterIonInfo('TMTpro-131C', 304.207100, 131.144500, 0.0),
    ReporterIonInfo('TMTpro-132N', 304.207100, 132.141535, 0.0),
    ReporterIonInfo('TMTpro-132C', 304.207100, 132.147855, 0.0),
    ReporterIonInfo('TMTpro-133N', 304.207100, 133.144890, 0.0),
    ReporterIonInfo('TMTpro-133C', 304.207100, 133.151210, 0.0),
    ReporterIonInfo('TMTpro-134N', 304.207100, 134.148245, 0.0)
]


IBT10_INFO = [
    ReporterIonInfo("T-114", 1, 114.128),
    ReporterIonInfo("T-115N", 1, 115.125),
    ReporterIonInfo("T-116N", 1, 116.129),
    ReporterIonInfo("T-117N", 1, 117.132),
    ReporterIonInfo("T-118N", 1, 118.135),
    ReporterIonInfo("T-119", 1, 119.139),
    ReporterIonInfo("T-115C", 1, 115.132),
    ReporterIonInfo("T-116C", 1, 116.135),
    ReporterIonInfo("T-117C", 1, 117.138),
    ReporterIonInfo("T-118C", 1, 118.142)
]


class SignatureIonExtractor(SignatureIonDetector):
    """Extracts signal for a set of target ions from each scan.
    """
    def extract(self, peak_list, error_tolerance=2e-5):
        """Extract the intensity for each signature ion.

        Parameters
        ----------
        peak_list : peak set-like
            The peak set to search. May be any object compatible with
            :class:`~.PeakSetMethods`.
        error_tolerance : float
            The PPM error tolerance for peak mass coordinate matching.

        Returns
        -------
        channels: dict[str, float]
            A mapping from signature ion name to its intensity.
        """
        result = {}
        peak_list = PeakSetMethods(peak_list)
        if peak_list.is_deconvoluted:
            for sig in self.signature_ions:
                result[sig.name] = 0
                matches = peak_list.all_peaks_for(sig.neutral_mass, error_tolerance)
                for match in matches:
                    if match.charge in sig.charge_states:
                        result[sig.name] += match.intensity
        else:
            for sig in self.signature_ions:
                result[sig.name] = 0
                for mz in sig.iter_mz():
                    match, error = peak_list.get_nearest_peak(mz)
                    if abs(error) / mz < error_tolerance:
                        result[sig.name] += match.intensity
        return result


    def __call__(self, peak_list, error_tolerance=2e-5):
        return self.extract(peak_list, error_tolerance)


class TMTReporterExtractor(SignatureIonExtractor):
    """A :class:`~.SignatureIonExtractor` for TMT reporter ions.
    """
    TMT_REAGENTS = {
        "tmt16": TMT16_INFO,
        "tmt11": TMT11_INFO,
        "tmt10": TMT10_INFO,
        "tmt2": TMT2_INFO,
        "tmt6": TMT6_INFO,
        "tmt0": TMT0_INFO,
        "ibt10": IBT10_INFO,
    }

    def _find_reagent(self, reagent):
        label = reagent.lower().replace("plex", "")
        if label in self.TMT_REAGENTS:
            reagents = self.TMT_REAGENTS[label]
        else:
            raise KeyError("Did not recognize reagent %s" % (reagent, ))
        return reagents

    def __init__(self, reagent="TMT11plex"):
        reagents = self._find_reagent(reagent)
        super(TMTReporterExtractor, self).__init__(
            [Target(
                n.name,
                n.reporter_mass_hcd)
                for n in reagents])



cdef class SpanningMixin(object):
    cdef:
        public double start
        public double end

    cdef bint _contains(self, double i)

    cpdef bint contains(self, double i)

    cpdef bint overlaps(self, SpanningMixin interval)

    cpdef double overlap_size(self, SpanningMixin interval)

    cpdef bint is_contained_in_interval(self, SpanningMixin interval)

    cpdef bint contains_interval(self, SpanningMixin interval)


cdef class Interval(SpanningMixin):
    cdef:
        public object members
        public dict data

cdef class SimpleInterval(SpanningMixin):
    @staticmethod
    cdef SimpleInterval _create(double start, double end)

cdef class IntervalTreeNode(object):
    cdef:
        public double center
        public double start
        public double end
        public IntervalTreeNode left
        public IntervalTreeNode right
        public list contained
        public int level
        public IntervalTreeNode parent

    cpdef list contains_point(self, double x)
    cpdef list _overlaps_interval(self, double start, double end)
    cpdef list overlaps(self, double start, double end)
    cpdef bint _eq(self, IntervalTreeNode other)


cdef class SpanningMixin2D(object):
    cdef:
        public double[2] _start
        public double[2] _end

    cdef bint _contains(self, double[2] i)
    cpdef bint contains(self, double[:] i)
    cpdef bint overlaps(self, SpanningMixin2D interval)
    cpdef bint contains_interval(self, SpanningMixin2D interval)


cdef class Interval2D(SpanningMixin2D):
    cdef:
        public list members
        public dict data


cdef class IntervalTreeNode2D(IntervalTreeNode):
    cdef:
        public object inner_organizer
        public IntervalTreeNode organized

    cpdef list _overlaps_interval_2d(self, double[::1] starts, double[::1] ends)
    cpdef list overlaps_2d(self, double[::1] starts, double[::1] ends)



from ms_peak_picker._c.peak_set cimport FittedPeak, PeakSet

from ms_deisotope._c.scoring cimport IsotopicFitRecord
from ms_deisotope._c.peak_dependency_network.intervals cimport SpanningMixin, IntervalTreeNode
from ms_deisotope._c.peak_dependency_network.subgraph cimport ConnectedSubgraph


cdef class PeakNode(object):
    """
    Represent a single FittedPeak and the junction of multiple
    IsotopicFitRecords which depend uponit

    Attributes
    ----------
    links : dict
        Mapping IsotopicFitRecords which depend upon this peak to those fit's scores
    peak : ms_peak_picker.FittedPeak
        The peak being depended upon
    """

    cdef:
        public FittedPeak peak
        public dict links
        long _hash

    @staticmethod
    cdef PeakNode _create(FittedPeak peak)


cdef class DependenceCluster(SpanningMixin):
    """
    Represent a cluster of isotopic fits which are overlapping

    Attributes
    ----------
    dependencies : set
        The set of all fits which are codependent
    start: float
        The smallest mz of all elements of this cluster
    end: float
        The largest mz of all elements of this cluster
    maximize: bool
        Whether the objective is to maximize or minimize the
        isotopic fit score
    """

    cdef:
        public object parent
        public list dependencies
        public bint maximize
        public IsotopicFitRecord best_fit

    cpdef _reset(self)
    cpdef add(self, IsotopicFitRecord fit)
    cpdef disjoint_subset(self)
    cpdef ConnectedSubgraph build_graph(self)
    cdef IsotopicFitRecord _best_fit(self)
    cpdef list disjoint_best_fits(self)
    cdef double _start(self)
    cdef double _end(self)


cdef class PeakDependenceGraphBase(object):
    cdef:
        public dict nodes
        public set dependencies
        public PeakSet peaklist
        public int max_missed_peaks
        public bint use_monoisotopic_superceded_filtering
        public bint maximize

        public list clusters
        public dict _solution_map
        public list _all_clusters

        public IntervalTreeNode _interval_tree

    cpdef reset(self)
    cpdef _populate_initial_graph(self)
    cpdef add_fit_dependence(self, IsotopicFitRecord fit_record)
    cpdef list nodes_for(self, IsotopicFitRecord fit_record, dict cache=*)
    cpdef drop_fit_dependence(self, IsotopicFitRecord fit_record)
    cpdef best_exact_fits(self)
    cpdef _gather_independent_clusters(self, dict nodes_for_cache=*)


import ast
import itertools
import json
import multiprocessing
import os
import re
import sys
import warnings

import click

from brainpy import periodic_table

from ms_deisotope.averagine import (
    Averagine, glycan as n_glycan_averagine, permethylated_glycan,
    peptide, glycopeptide, heparin, heparan_sulfate)

from ms_deisotope.task.log_utils import fmt_msg


def processes_option(f):
    opt = click.option(
        "-p", "--processes", 'processes', type=click.IntRange(1, multiprocessing.cpu_count()),
        default=min(multiprocessing.cpu_count(), 4), help=('Number of worker processes to use. Defaults to 4 '
                                                           'or the number of CPUs, whichever is lower'))
    return opt(f)


def validate_element(element):
    valid = element in periodic_table
    if not valid:
        raise click.Abort("%r is not a valid element" % element)
    return valid


def parse_averagine_formula(formula):
    if isinstance(formula, Averagine):
        return formula
    return Averagine({k: float(v) for k, v in re.findall(r"([A-Z][a-z]*)([0-9\.]*)", formula)
                      if float(v or 0) > 0 and validate_element(k)})


averagines = {
    'glycan': n_glycan_averagine,
    'permethylated-glycan': permethylated_glycan,
    'peptide': peptide,
    'glycopeptide': glycopeptide,
    'heparin': heparin,
    "heparan-sulfate": heparan_sulfate
}


def validate_averagine(averagine_string):
    if isinstance(averagine_string, Averagine):
        return averagine_string
    if averagine_string in averagines:
        return averagines[averagine_string]
    else:
        return parse_averagine_formula(averagine_string)


class AveragineParamType(click.types.StringParamType):
    name = "MODEL"

    models = averagines

    def convert(self, value, param, ctx):
        return validate_averagine(value)

    def get_metavar(self, param):
        return '[%s]' % '|'.join(sorted(averagines.keys()))

    def get_missing_message(self, param):
        return 'Choose from %s, or provide a formula.' % ', '.join(self.choices)


def register_debug_hook():
    import traceback

    def info(type, value, tb):
        if hasattr(sys, 'ps1') or not sys.stderr.isatty():
            sys.__excepthook__(type, value, tb)
        else:
            try:
                import ipdb as pdb_api
            except ImportError:
                import pdb as pdb_api
            traceback.print_exception(type, value, tb)
            pdb_api.post_mortem(tb)

    sys.excepthook = info
    # logging.basicConfig(level="DEBUG")


def envar_to_bool(env_val: str) -> bool:
    env_val = env_val.lower()
    if not env_val:
        return False
    if env_val in ('0', 'no', 'false', 'off'):
        return False
    elif env_val in ('1', 'yes', 'true', 'on'):
        return True
    return None


def is_debug_mode() -> bool:
    env_val = os.environ.get('MS_DEISOTOPE_DEBUG', '').lower()
    val = envar_to_bool(env_val)
    if val is not None:
        return val
    else:
        warnings.warn("MS_DEISOTOPE_DEBUG value %r was not recognized. Enabling debug mode" % (env_val, ))
        return True


class Spinner(object):
    """Super-simple synchronous CLI spinner implementation.
    """
    _default_symbols = ['-', '/', '|', '\\']

    def __init__(self, stream=None, symbols=None, title=None):
        if symbols is None:
            symbols = list(self._default_symbols)
        if stream is None:
            stream = click.get_text_stream('stdout')
        self.symbol_cycle = itertools.cycle(symbols)
        self.stream = stream
        self.title = title

    def _next_symbol(self):
        symbol = next(self.symbol_cycle)
        return symbol

    def update(self, *args, **kwargs):
        if self.stream.isatty():
            return self._update(*args, **kwargs)

    def _update(self, *args, **kwargs):
        self.stream.write("\b")
        self.stream.flush()
        symbol = self._next_symbol()
        self.stream.write(symbol)
        self.stream.flush()

    def __enter__(self):
        if self.title is not None:
            self.stream.write(self.title)
            self.stream.write("  ")
            self.stream.flush()
        return self

    def __exit__(self, *args):
        self.stream.write("\n")
        self.stream.flush()

    def __repr__(self):
        template = "{self.__class__.__name__}({self.stream}, {self.symbol_cycle})"
        return template.format(self=self)

spinner = Spinner


class ProgressLogger(object):
    """A simple text logger that wraps an iterable and logs update messages as chunks are requested.

    This class tries to emulate :func:`click.progressbar` for use when the attached STDERR stream is not
    a terminal.
    """
    def __init__(self, iterable=None, length=None, label=None, item_show_func=None, interval=None,
                 file=None, writer=None, **kwargs):
        if iterable is not None:
            try:
                length = len(iterable)
            except TypeError:
                pass
        if interval is None:
            if length is None:
                interval = 1000
            else:
                interval = int(length * 0.05)
        if interval == 0:
            interval = 1
        if label is None:
            label = ''
        if file is None:
            file = click.get_text_stream('stderr')
        self.iterable = iterable
        self.length = length
        self.item_show_func = item_show_func
        self.current_item = None
        self.count = 0
        self.interval = interval
        self.last_update = 0
        self.label = label
        self.file = file
        self.writer = writer

    def update(self, n):
        self.count += n
        if self.count > self.last_update:
            self._log()
            self.last_update += self.interval * (int(n // self.interval) + 1)

    def _log(self):
        item = self.current_item
        if self.item_show_func is not None:
            show = self.item_show_func(item)
        else:
            show = str(item)
        i = self.count
        label = self.label
        if self.length is not None and self.length != 0:
            prog_label = "%s %d/%d (%0.2f%%)" % (label, i, self.length, i * 100.0 / self.length)
        else:
            prog_label = "%s %d" % (label, i)
        message = fmt_msg("%s: %s" % (prog_label, show))
        if self.writer is None:
            click.echo(message, file=self.file, err=True)
        else:
            self.writer(message)

    def __iter__(self):
        for item in self.iterable:
            self.current_item = item
            self.update(1)
            yield item

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass


def progress(*args, **kwargs):
    """A wrapper that will dispatch to :func:`click.progressbar` when `sys.stdout` is a
    TTY and :class:`ProgressLogger` otherwise.
    """
    if sys.stdout.isatty():
        return click.progressbar(*args, **kwargs)
    else:
        return ProgressLogger(*args, **kwargs)


def progress_iter(*args, **kwargs):
    prog = progress(*args, **kwargs)
    with prog:
        yield from prog


def type_cast(value):
    try:
        return json.loads(value)
    except (TypeError, ValueError):
        try:
            return ast.literal_eval(value)
        except (TypeError, ValueError):
            return str(value)



from ms_deisotope._c.feature_map.lcms_feature cimport FeatureBase, LCMSFeatureTreeNode, LCMSFeature
from ms_deisotope._c.peak_set cimport DeconvolutedPeak, Envelope
from ms_deisotope._c.averagine cimport TheoreticalIsotopicPattern

cimport numpy as np
import numpy as np

cdef class map_coord:
    cdef:
        public double mz
        public double time

    cdef bint ge(self, map_coord other)
    cdef bint le(self, map_coord other)

    @staticmethod
    cdef map_coord _create(double mz, double time)

    cpdef map_coord copy(self)


cdef class LCMSFeatureSetFit(object):
    cdef:
        public list features
        public TheoreticalIsotopicPattern theoretical
        public list supporters
        public double score
        public double mz
        public double neutral_mass
        public int charge
        public object data
        public size_t missing_features
        public size_t n_points
        public FeatureBase monoisotopic_feature
        public object scores
        public object times

    cpdef bint _eq(self, LCMSFeatureSetFit other)
    cpdef bint _ne(self, LCMSFeatureSetFit other)
    cpdef bint _lt(self, LCMSFeatureSetFit other)
    cpdef bint _gt(self, LCMSFeatureSetFit other)

    cpdef int count_null_features(self)
    cpdef bint has_multiple_real_features(self)

    cdef map_coord get_start(self)
    cdef map_coord get_end(self)

    @staticmethod
    cdef LCMSFeatureSetFit _create(list features, TheoreticalIsotopicPattern theoretical,
                                   double score, int charge, size_t missing_features,
                                   list supporters, object data, double neutral_mass,
                                   size_t n_points, object scores, object times)


cpdef Envelope _sum_envelopes(LCMSFeature self)


from __future__ import print_function

import logging
import logging.handlers
import multiprocessing
import sys
import threading
import traceback
import re

from typing import Dict

from datetime import datetime

from queue import Empty


logger = logging.getLogger("ms_deisotope.task")


def ensure_text(obj):
    return str(obj)


def fmt_msg(*message):
    return u"%s %s" % (ensure_text(datetime.now().isoformat(' ')), u', '.join(map(ensure_text, message)))


def printer(obj, *message):
    print(fmt_msg(*message))


def show_message(*message):
    print(fmt_msg(*message))


def debug_printer(obj, *message):
    if obj.in_debug_mode():
        print(u"DEBUG:" + fmt_msg(*message))


class CallInterval(object):
    """Call a function every `interval` seconds from
    a separate thread.

    Attributes
    ----------
    stopped: threading.Event
        A semaphore lock that controls when to run `call_target`
    call_target: callable
        The thing to call every `interval` seconds
    args: iterable
        Arguments for `call_target`
    interval: number
        Time between calls to `call_target`
    """

    def __init__(self, interval, call_target, *args):
        self.stopped = threading.Event()
        self.interval = interval
        self.call_target = call_target
        self.args = args
        self.thread = threading.Thread(target=self.mainloop)
        self.thread.daemon = True

    def mainloop(self):
        while not self.stopped.wait(self.interval):
            try:
                self.call_target(*self.args)
            except (KeyboardInterrupt, SystemExit):
                self.stop()
            except Exception as e:
                logger.exception("An error occurred in %r", self, exc_info=e)

    def start(self):
        self.thread.start()

    def stop(self):
        self.stopped.set()


class MessageSpooler(object):
    """An IPC-based logging helper

    Attributes
    ----------
    halting : bool
        Whether the object is attempting to
        stop, so that the internal thread can
        tell when it should stop and tell other
        objects using it it is trying to stop
    handler : Callable
        A Callable object which can be used to do
        the actual logging
    message_queue : multiprocessing.Queue
        The Inter-Process Communication queue
    thread : threading.Thread
        The internal listener thread that will consume
        message_queue work items
    """
    def __init__(self, handler):
        self.handler = handler
        self.message_queue = multiprocessing.Queue()
        self.halting = False
        self.thread = threading.Thread(target=self.run)
        self.thread.daemon = True
        self.thread.start()

    def run(self):
        while not self.halting:
            try:
                message = self.message_queue.get(True, 2)
                self.handler(*message)
            except Empty:
                continue

    def stop(self):
        self.halting = True
        self.thread.join()

    def sender(self):
        return MessageSender(self.message_queue)


class MessageSender(object):
    """A simple callable for pushing objects into an IPC
    queue.

    Attributes
    ----------
    queue : multiprocessing.Queue
        The Inter-Process Communication queue
    """
    def __init__(self, queue):
        self.queue = queue

    def __call__(self, *message):
        self.send(*message)

    def send(self, *message):
        self.queue.put(message)


class LogUtilsMixin(object):

    logger_state = None
    print_fn = printer
    debug_print_fn = debug_printer
    error_print_fn = printer
    warn_print_fn = printer

    _debug_enabled = None

    @classmethod
    def log_with_logger(cls, logger):
        cls.logger_state = logger
        cls.print_fn = logger.info
        cls.debug_print_fn = logger.debug
        cls.error_print_fn = logger.error
        cls.warn_print_fn = logger.warn

    def pipe_from(self, task):
        task.print_fn = self.log
        task.debug_print_fn = self.debug
        task.error_print_fn = self.error
        task.warn_print_fn = self.warn

    def in_debug_mode(self):
        if self._debug_enabled is None:
            logger_state = self.logger_state
            if logger_state is not None:
                self._debug_enabled = logger_state.isEnabledFor("DEBUG")
        return bool(self._debug_enabled)

    def log(self, *message):
        self.print_fn(u', '.join(map(ensure_text, message)))

    def warn(self, *message):
        self.warn_print_fn(u', '.join(map(ensure_text, message)))

    def debug(self, *message):
        self.debug_print_fn(u', '.join(map(ensure_text, message)))

    def error(self, *message, **kwargs):
        exception = kwargs.get("exception")
        self.error_print_fn(u', '.join(map(ensure_text, message)))
        if exception is not None:
            self.error_print_fn(traceback.format_exc(exception))

    def ipc_logger(self, handler=None):
        if handler is None:
            def default_handler(message):
                self.log(message)
            handler = default_handler
        return MessageSpooler(handler)


class ProgressUpdater(LogUtilsMixin):
    def __init__(self, *args, **kwargs):
        kwargs['item_show_func'] = self._prepare_message
        kwargs.setdefault("color", True)
        kwargs.setdefault("fill_char", click.style('-', 'blue'))
        try:
            import click
            self.progress_bar = click.progressbar(*args, **kwargs)
        except ImportError:
            self.progress_bar = None

    def _prepare_message(self, *message):
        if len(message) == 1:
            if message[0] is None:
                return ""
        return ', '.join(map(str, message))

    def log(self, *message):
        if self.progress_bar is not None:
            if not self.progress_bar.is_hidden:
                self.progress_bar.current_item = self._prepare_message(
                    *message)
                return
        super(ProgressUpdater, self).log(*message)

    def warn(self, *message):
        if self.progress_bar is not None:
            if not self.progress_bar.is_hidden:
                self.progress_bar.current_item = "WARN: " + self._prepare_message(
                    *message)
                return
        super(ProgressUpdater, self).warn(*message)

    def debug(self, *message):
        if self.progress_bar is not None:
            if not self.progress_bar.is_hidden:
                self.progress_bar.current_item = "DEBUG: " + self._prepare_message(
                    *message)
                return
        super(ProgressUpdater, self).debug(*message)

    def begin(self):
        if self.progress_bar is not None:
            self.progress_bar.__enter__()

    def end(self, exc_type=None, exc_value=None, tb=None):
        if self.progress_bar is not None:
            self.progress_bar.__exit__(exc_type, exc_value, tb)

    def __enter__(self):
        return self.begin()

    def __exit__(self, exc_type, exc_value, tb):
        return self.end(exc_type, exc_value, tb)

    def update(self, i, *message):
        if self.progress_bar is not None:
            self.progress_bar.update(i)
        if message:
            self.log(*message)


class ProcessAwareFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        d = record.__dict__
        try:
            if d['processName'] == "MainProcess":
                d['maybeproc'] = ''
            else:
                d['maybeproc'] = ":%s:" % d['processName'].replace(
                    "Process", '')
        except KeyError:
            d['maybeproc'] = ''
        return super(ProcessAwareFormatter, self).format(record)


class LevelAwareColoredLogFormatter(ProcessAwareFormatter):
    try:
        from colorama import Fore, Style
        # GREY = Fore.WHITE
        GREY = ''
        BLUE = Fore.BLUE
        GREEN = Fore.GREEN
        YELLOW = Fore.YELLOW
        RED = Fore.RED
        BRIGHT = Style.BRIGHT
        DIM = Style.DIM
        BOLD_RED = Fore.RED + Style.BRIGHT
        RESET = Style.RESET_ALL
    except ImportError:
        GREY = ''
        BLUE = ''
        GREEN = ''
        YELLOW = ''
        RED = ''
        BRIGHT = ''
        DIM = ''
        BOLD_RED = ''
        RESET = ''

    def _colorize_field(self, fmt: str, field: str, color: str) -> str:
        return re.sub("(" + field + ")", color + r"\1" + self.RESET, fmt)


    def _patch_fmt(self, fmt: str, level_color: str) -> str:
        fmt = self._colorize_field(fmt, r"%\(asctime\)s", self.GREEN)
        fmt = self._colorize_field(fmt, r"%\(name\).*?s", self.BLUE)
        fmt = self._colorize_field(fmt, r"%\(message\).*?s", self.GREY)
        if level_color:
            fmt = self._colorize_field(fmt, r"%\(levelname\).*?s", level_color)
        return fmt

    def __init__(self, fmt, level_color=None, **kwargs):
        fmt = self._patch_fmt(fmt, level_color=level_color)
        super().__init__(fmt, **kwargs)


class ColoringFormatter(logging.Formatter):
    level_to_color = {
        logging.INFO: LevelAwareColoredLogFormatter.GREEN,
        logging.DEBUG: LevelAwareColoredLogFormatter.GREY + LevelAwareColoredLogFormatter.DIM,
        logging.WARN: LevelAwareColoredLogFormatter.YELLOW + LevelAwareColoredLogFormatter.BRIGHT,
        logging.ERROR: LevelAwareColoredLogFormatter.BOLD_RED,
        logging.CRITICAL: LevelAwareColoredLogFormatter.BOLD_RED,
        logging.FATAL: LevelAwareColoredLogFormatter.RED + LevelAwareColoredLogFormatter.DIM,
    }

    _formatters: Dict[int, LevelAwareColoredLogFormatter]

    def __init__(self, fmt: str, **kwargs):
        self._formatters = {}
        for level, style in self.level_to_color.items():
            self._formatters[level] = LevelAwareColoredLogFormatter(fmt, level_color=style, **kwargs)

    def format(self, record: logging.LogRecord) -> str:
        fmtr = self._formatters[record.levelno]
        return fmtr.format(record)



def init_logging(filename=None, queue=None):
    logging.basicConfig(
        level="INFO", format='[%(asctime)s] %(levelname)s: %(message)s',
        datefmt="%H:%M:%S", handlers=[]
    )
    logging.captureWarnings(True)

    logger = logging.getLogger('ms_deisotope')
    format_string = '[%(asctime)s] %(levelname).1s | %(name)s | %(message)s'
    formatter = ProcessAwareFormatter(format_string, datefmt="%H:%M:%S")
    colorized_formatter = ColoringFormatter(format_string, datefmt="%H:%M:%S")

    # If there was a queue, don't add any other handlers, route all logging through
    # the queue
    if queue:
        queue_handler = logging.handlers.QueueHandler(queue)
        queue_handler.setFormatter(formatter)
        logger.addHandler(queue_handler)
    else:
        # Otherwise, configure handlers for `ms_deisotope`
        stderr_handler = logging.StreamHandler(sys.stderr)
        if sys.stderr.isatty():
            stderr_handler.setFormatter(colorized_formatter)
        else:
            stderr_handler.setFormatter(formatter)
        logger.addHandler(stderr_handler)
        if filename:
            file_handler = logging.FileHandler(filename=filename, mode='w', encoding='utf8')
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)
    LogUtilsMixin.log_with_logger(logger)



cimport cython
from cpython.list cimport PyList_GetItem, PyList_Size, PyList_AsTuple
from cpython.tuple cimport PyTuple_GetItem, PyTuple_Size
from cpython.set cimport PySet_Add

from ms_deisotope._c.peak_dependency_network.intervals cimport (SpanningMixin, Interval)
from ms_deisotope._c.scoring cimport IsotopicFitRecord

from ms_peak_picker._c.peak_set cimport FittedPeak


@cython.final
cdef class FitNode(SpanningMixin):
    def __init__(self, IsotopicFitRecord fit, size_t index=-1):
        self.index = index
        self.fit = fit
        self.edges = set()
        self.overlap_edges = set()
        self.peak_indices = set()
        self._hash = hash(fit)
        self.score = fit.score
        self._init_fields(fit.experimental)

    @staticmethod
    cdef FitNode _create(IsotopicFitRecord fit, size_t index):
        cdef:
            FitNode node
        node = FitNode.__new__(FitNode)
        node.index = index
        node.fit = fit
        node.edges = set()
        node.overlap_edges = set()
        node.peak_indices = set()
        node._hash = hash(fit)
        node.score = fit.score
        node._init_fields(fit.experimental)
        return node

    cdef void _init_fields(self, list experimental):
        cdef:
            size_t i, n
            FittedPeak peak
        n = PyList_Size(experimental)
        for i in range(n):
            # Placeholder Peaks have a peak_count of -1
            peak = <FittedPeak>PyList_GetItem(experimental, i)
            if peak.peak_count >= 0:
                PySet_Add(self.peak_indices, peak)
        if n > 0:
            self.end = peak.mz
            self.start = (<FittedPeak>PyList_GetItem(experimental, 0)).mz
        else:
            self.end = 0
            self.start = 0

    def __hash__(self):
        return self._hash

    def __eq__(self, FitNode other):
        return self.fit is other.fit

    def __gt__(self, FitNode other):
        return self.score > other.score

    def __lt__(self, FitNode other):
        return self.score < other.score

    def __ne__(self, FitNode other):
        return self.fit is not other.fit

    cpdef bint isdisjoint(self, FitNode other):
        return self.peak_indices.isdisjoint(other.peak_indices)

    cpdef visit(self, FitNode other):
        if self.isdisjoint(other):
            PySet_Add(self.edges, other)
            PySet_Add(other.edges, self)
        else:
            PySet_Add(self.overlap_edges, other)
            PySet_Add(other.overlap_edges, self)

    def __repr__(self):
        return "FitNode(%r)" % (self.fit, )


cpdef bint span_overlap(FitNode a, FitNode b):
    return a.__contains__(b.start) or a.__contains__(b.end - 1) or\
        b.__contains__(a.start) or b.__contains__(a.end - 1)


cpdef bint peak_overlap(FitNode a, FitNode b):
    return len(a.peak_indices & b.peak_indices) > 0


@cython.nonecheck(False)
cpdef list layout_layers(list envelopes, bint maximize=True):
    """
    Produce a non-overlapping stacked layout of individual envelopes.
    """
    cdef:
        list layers, layer
        size_t env_i, env_n, layer_i, layer_n, member_i, member_n
        bint placed, collision
        FitNode envelope, member

    layers = [[]]
    envelopes.sort(reverse=maximize)
    env_n = PyList_Size(envelopes)
    for env_i in range(env_n):
        envelope = <FitNode>PyList_GetItem(envelopes, env_i)
        placed = False
        layer_n = PyList_Size(layers)
        for layer_i in range(layer_n):
            layer = <list>PyList_GetItem(layers, layer_i)
            collision = False
            member_n = PyList_Size(layer)
            for member_i in range(member_n):
                member = <FitNode>PyList_GetItem(layer, member_i)
                if peak_overlap(envelope, member):
                    collision = True
                    break
            if not collision:
                layer.append(envelope)
                placed = True
                break
        if not placed:
            layers.append([envelope])
    return layers


@cython.final
@cython.freelist(5)
cdef class GreedySubgraphSelection(object):
    cdef:
        public list intervals
        public list _layers
        public bint maximize

    def __init__(self, object subgraph, bint maximize=True):
        self.intervals = list(subgraph)
        self._layers = [[]]
        self.maximize = maximize
        self._build_layers()

    cdef void _build_layers(self):
        self._layers = layout_layers(self.intervals, maximize=self.maximize)

    cdef list _select_best_subset(self):
        return self._layers[0]

    cpdef select(self):
        return self._select_best_subset()


@cython.freelist(5)
cdef class ConnectedSubgraph(object):
    def __init__(self, object fits, bint maximize=True):
        self.maximize = maximize
        self._init_nodes(list(fits))
        self.populate_edges()

    cdef void _init_nodes(self, list fits):
        cdef:
            list nodes
            size_t i, n
            IsotopicFitRecord fit
        nodes = list()
        n = PyList_Size(fits)
        for i in range(n):
            fit = <IsotopicFitRecord>PyList_GetItem(fits, i)
            nodes.append(FitNode._create(fit, i))
        self.nodes = PyList_AsTuple(nodes)

    cdef void populate_edges(self):
        cdef:
            size_t i, j, n
            FitNode node, other

        n = PyTuple_Size(self.nodes)

        for i in range(n):
            node = <FitNode>PyTuple_GetItem(self.nodes, i)
            for j in range(n):
                if i == j:
                    continue
                other = <FitNode>PyTuple_GetItem(self.nodes, j)
                node.visit(other)

    def __getitem__(self, i):
        return self.nodes[i]

    def __iter__(self):
        return iter(self.nodes)

    def __len__(self):
        return len(self.nodes)

    cpdef find_heaviest_path(self, method="greedy"):
        cdef:
            GreedySubgraphSelection solution
        if len(self) == 1:
            return set(self.nodes)
        if method == "greedy":
            solution = GreedySubgraphSelection(self, maximize=self.maximize)
            return solution.select()
        else:
            raise NotImplementedError(method)



import operator
import math
from collections import namedtuple
from typing import Any, List, Optional, Sequence, TYPE_CHECKING

from brainpy import mass_charge_ratio, neutral_mass as calc_neutral_mass, PROTON

from ms_peak_picker import simple_peak, PeakSet as FittedPeakSet
from ms_deisotope.peak_dependency_network.intervals import SimpleInterval

from ms_deisotope.scoring import IsotopicFitRecord

from .utils import Base, ppm_error


class _Index(object):
    """
    Stores the ordered index of an object under sort by `mz` and
    sort by `neutral_mass`

    Attributes
    ----------
    mz : int
        Index of this object (or its parent), ordered by mz
    neutral_mass : int
        Index of this object (or its parent), ordered by neutral mass
    """

    __slots__ = ["neutral_mass", "mz"]

    neutral_mass: int
    mz: int

    def __init__(self, neutral_mass=None, mz=None):
        self.neutral_mass = neutral_mass
        self.mz = mz

    def __reduce__(self):
        return self.__class__, (self.neutral_mass, self.mz)

    def __repr__(self):
        return "%d|%d" % (self.neutral_mass, self.mz)

    def clone(self):
        return self.__class__(self.neutral_mass, self.mz)


EnvelopePair = namedtuple("EnvelopePair", ("mz", "intensity"))


class Envelope(Sequence[EnvelopePair]):
    """
    Represents a sequence of (mz, intensity) pairs which store peak positions
    that were matched. Since these peaks may later be mutated by subtraction,
    this structure stores copies of the numbers

    Supports the Sequence protocol, and is read-only.

    Attributes
    ----------
    pairs : tuple of tuple of (float, float)
        list of (mz, intensity) pairs
    """

    __slots__ = ['pairs']

    def __init__(self, pairs):
        self.pairs = [EnvelopePair(*x) for x in pairs]

    def __reduce__(self):
        return self.__class__, (self.pairs,)

    def __getitem__(self, i):
        return self.pairs[i]

    def __iter__(self):
        return iter(self.pairs)

    def __len__(self):
        return len(self.pairs)

    def __repr__(self):
        return "[%s]" % (', '.join("(%0.4f, %0.2f)" % t for t in self),)

    def clone(self):
        return self.__class__(self)


class DeconvolutedPeak(Base):
    """Represent a single deconvoluted peak which represents an aggregated isotopic
    pattern collapsed to its monoisotopic peak, with a known charge state

    Attributes
    ----------
    a_to_a2_ratio : float
        Ratio of intensities of A peak to A+2 peak
    average_mass : float
        The averaged neutral mass of the composition, the weighted average
        of the envelope peaks.
    charge : int
        The signed charge state of the isotopic pattern
    envelope : Envelope
        The sequence of (mz, intensity) pairs which map to this peak
    full_width_at_half_max : float
        The averaged full width at half max of this isotopic pattern
    index : _Index
        The position of this peak in different orderings
    intensity : float
        The summed height of the peaks in the isotopic pattern this peak captures
    most_abundant_mass : float
        The neutral mass of the most abundant peak in the isotopic pattern this peak captures
    mz : float
        The mass-charge-ratio of the monoisotopic peak
    neutral_mass : float
        The neutral mass of the monoisotopic peak
    score : float
        An assigned value describing the quality of this peak's fit. The semantics of this score
        depends upon the scoring function
    signal_to_noise : float
        The average signal-to-noise ratio of the peaks in the isotopic pattern this peak captures
    """

    __slots__ = [
        "neutral_mass", "intensity", "signal_to_noise",
        "index", "full_width_at_half_max", "charge",
        "a_to_a2_ratio", "most_abundant_mass", "average_mass",
        "score", "envelope", "mz", "fit", "chosen_for_msms", 'area'
    ]

    neutral_mass: float
    intensity: float
    signal_to_noise: float
    index: _Index
    full_width_at_half_max: float
    charge: int
    a_to_a2_ratio: float
    most_abundant_mass: float
    average_mass: float
    score: float
    envelope: Envelope
    mz: float
    fit: Optional[IsotopicFitRecord]
    chosen_for_msms: bool
    area: float

    def __init__(self, neutral_mass, intensity, charge, signal_to_noise, index, full_width_at_half_max,
                 a_to_a2_ratio=0, most_abundant_mass=0, average_mass=0, score=0,
                 envelope=None, mz=0, fit=None, chosen_for_msms=False, area=0):
        if index is None:
            index = _Index()
        if envelope is None:
            envelope = ()
        self.neutral_mass = neutral_mass
        self.intensity = intensity
        self.signal_to_noise = signal_to_noise
        self.index = index
        self.full_width_at_half_max = full_width_at_half_max
        self.charge = charge
        self.a_to_a2_ratio = a_to_a2_ratio
        self.most_abundant_mass = most_abundant_mass
        self.average_mass = average_mass
        self.score = score
        self.envelope = Envelope(envelope)
        self.mz = mz or mass_charge_ratio(self.neutral_mass, self.charge)
        self.fit = fit
        self.chosen_for_msms = chosen_for_msms
        self.area = area

    def __eq__(self, other):
        epsilon = 1e-3
        return (abs(self.neutral_mass - other.neutral_mass) < epsilon) and (
            abs(self.intensity - other.intensity) < epsilon) and (
            self.charge == other.charge) and (
            abs(self.score - other.score) < epsilon)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.mz, self.intensity, self.charge))

    def clone(self):
        return DeconvolutedPeak(self.neutral_mass, self.intensity, self.charge, self.signal_to_noise,
                                self.index, self.full_width_at_half_max, self.a_to_a2_ratio,
                                self.most_abundant_mass, self.average_mass, self.score,
                                self.envelope, self.mz, self.fit, self.chosen_for_msms,
                                self.area)

    def __reduce__(self):
        return DeconvolutedPeak, (self.neutral_mass, self.intensity, self.charge, self.signal_to_noise,
                                  self.index, self.full_width_at_half_max, self.a_to_a2_ratio,
                                  self.most_abundant_mass, self.average_mass, self.score,
                                  self.envelope, self.mz, self.fit, self.chosen_for_msms, self.area)


class IonMobilityDeconvolutedPeak(DeconvolutedPeak):
    drift_time: float

    def __init__(self, neutral_mass, intensity, charge, signal_to_noise, index, full_width_at_half_max,
                 a_to_a2_ratio=0, most_abundant_mass=0, average_mass=0, score=0,
                 envelope=(), mz=0, fit=None, chosen_for_msms=False, area=0, drift_time=0):
        super(IonMobilityDeconvolutedPeak, self).__init__(
            neutral_mass, intensity, charge, signal_to_noise, index, full_width_at_half_max,
            a_to_a2_ratio, most_abundant_mass, average_mass, score,
            envelope, mz, fit, chosen_for_msms, area)
        self.drift_time = drift_time

    def clone(self):
        return IonMobilityDeconvolutedPeak(
            self.neutral_mass, self.intensity, self.charge, self.signal_to_noise,
            self.index, self.full_width_at_half_max, self.a_to_a2_ratio,
            self.most_abundant_mass, self.average_mass, self.score,
            self.envelope, self.mz, self.fit, self.chosen_for_msms, self.area,
            self.drift_time)

    def __reduce__(self):
        return IonMobilityDeconvolutedPeak, (
            self.neutral_mass, self.intensity, self.charge, self.signal_to_noise,
            self.index, self.full_width_at_half_max, self.a_to_a2_ratio,
            self.most_abundant_mass, self.average_mass, self.score,
            self.envelope, self.mz, self.fit, self.chosen_for_msms, self.area,
            self.drift_time)

    def __repr__(self):
        return ("{self.__class__.__name__}(a_to_a2_ratio={self.a_to_a2_ratio}, area={self.area}, "
                "average_mass={self.average_mass}, charge={self.charge}, chosen_for_msms={self.chosen_for_msms}, "
                "envelope={self.envelope}, full_width_at_half_max={self.full_width_at_half_max}, index={self.index}, "
                "intensity={self.intensity}, most_abundant_mass={self.most_abundant_mass}, mz={self.mz}, "
                "neutral_mass={self.neutral_mass}, score={self.score}, signal_to_noise={self.signal_to_noise},"
                "drift_time={self.drift_time})").format(self=self)


class DeconvolutedPeakSolution(DeconvolutedPeak):
    """
    Extends :class:`DeconvolutedPeak` to also include a reference to
    the :class:`IsotopicFitRecord` instance it is derived from, and optionally an
    object which the fit derives from.

    Attributes
    ----------
    fit : IsotpicFitRecord
        The isotopic pattern fit used to construct this peak
    solution : object
        Representation of the "source" of the isotopic fit
    """

    __slots__ = [
        "solution", "fit",
        "neutral_mass", "intensity", "signal_to_noise",
        "index", "full_width_at_half_max", "charge",
        "a_to_a2_ratio", "most_abundant_mass", "average_mass",
        "score", "envelope", "mz", "chosen_for_msms", 'area'
    ]

    solution: Any

    def __init__(self, solution, fit, *args, **kwargs):
        self.solution = solution
        super(DeconvolutedPeakSolution, self).__init__(*args, **kwargs)
        self.fit = fit

    def clone(self):
        return DeconvolutedPeakSolution(
            self.solution, self.fit, self.neutral_mass, self.intensity, self.charge, self.signal_to_noise,
            self.index, self.full_width_at_half_max, self.a_to_a2_ratio,
            self.most_abundant_mass, self.average_mass, self.score,
            self.envelope, self.mz, self.chosen_for_msms, self.area)

    def __reduce__(self):
        return DeconvolutedPeakSolution, (
            self.solution, self.fit, self.neutral_mass, self.intensity, self.charge, self.signal_to_noise,
            self.index, self.full_width_at_half_max, self.a_to_a2_ratio,
            self.most_abundant_mass, self.average_mass, self.score,
            self.envelope, self.mz, self.chosen_for_msms, self.area)

    def __iter__(self):
        yield self.solution
        yield self
        yield self.fit

    @classmethod
    def from_peak(cls, peak, solution):
        inst = cls(
            solution, peak.fit, peak.neutral_mass, peak.intensity,
            peak.charge, peak.signal_to_noise, peak.index, peak.full_width_at_half_max,
            peak.a_to_a2_ratio, peak.most_abundant_mass, peak.average_mass,
            peak.score, peak.envelope, peak.mz, peak.chosen_for_msms, peak.area)
        return inst


class DeconvolutedPeakSet(Base):
    """
    Represents a collection of :class:`DeconvolutedPeak` instances under multiple orderings.

    Attributes
    ----------
    peaks : tuple of DeconvolutedPeak
        Collection of peaks ordered by `neutral_mass`
    _mz_ordered: tuple of DeconvolutedPeak
        Collection of peaks ordered by `mz`
    """

    def __init__(self, peaks):
        self.peaks = peaks
        self._mz_ordered = None

    def reindex(self):
        """
        Updates the :attr:`index` of each peak in `self` and updates the
        sorted order.

        Returns
        -------
        self: DeconvolutedPeakSet
        """
        self.peaks = tuple(
            sorted(self.peaks, key=operator.attrgetter("neutral_mass")))
        self._mz_ordered = tuple(
            sorted(self.peaks, key=operator.attrgetter("mz")))
        for i, peak in enumerate(self.peaks):
            peak.index = _Index()
            peak.index.neutral_mass = i
        for i, peak in enumerate(self._mz_ordered):
            peak.index.mz = i
        return self

    def __len__(self):
        return len(self.peaks)

    def get_nearest_peak(self, neutral_mass, use_mz=False):
        if use_mz:
            return _get_nearest_peak(self._mz_ordered, neutral_mass, use_mz=use_mz)
        else:
            return _get_nearest_peak(self.peaks, neutral_mass, use_mz=use_mz)

    def has_peak(self, neutral_mass, tolerance=1e-5, use_mz=False):
        if use_mz:
            return binary_search(self._mz_ordered, neutral_mass, tolerance, mz_getter)
        return binary_search(self.peaks, neutral_mass, tolerance, neutral_mass_getter)

    def all_peaks_for(self, neutral_mass, tolerance=1e-5):
        lo = neutral_mass - neutral_mass * tolerance
        hi = neutral_mass + neutral_mass * tolerance
        lo_peak, lo_err = self.get_nearest_peak(lo)
        hi_peak, hi_err = self.get_nearest_peak(hi)
        lo_ix = lo_peak.index.neutral_mass
        if abs(ppm_error(lo_peak.neutral_mass, neutral_mass)) > tolerance:
            lo_ix += 1
        hi_ix = hi_peak.index.neutral_mass + 1
        if abs(ppm_error(hi_peak.neutral_mass, neutral_mass)) > tolerance:
            hi_ix -= 1
        return tuple(self[lo_ix:hi_ix])

    def __repr__(self):
        return "<DeconvolutedPeakSet %d Peaks>" % (len(self))

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self.__class__(tuple(p.clone() for p in self.peaks[item]))
        return self.peaks[item]

    def __iter__(self):
        return iter(self.peaks)

    def clone(self):
        return self.__class__(tuple(p.clone() for p in self))

    def copy(self):
        return self.clone()

    def __eq__(self, other):
        try:
            return self.peaks == other.peaks
        except AttributeError:
            return tuple(self) == tuple(other)

    def __ne__(self, other):
        try:
            return self.peaks != other.peaks
        except AttributeError:
            return tuple(self) != tuple(other)

    def between(self, m1, m2, tolerance=1e-5, use_mz=False):
        acc = []
        collecting = False
        if not use_mz:
            getter = operator.attrgetter("neutral_mass")
            iterator = self.peaks
        else:
            getter = operator.attrgetter("mz")
            iterator = self._mz_ordered
        for peak in iterator:
            if not collecting and getter(peak) >= m1:
                collecting = True
            elif collecting and getter(peak) > m2:
                break

            if collecting:
                acc.append(peak.clone())

        return self.__class__(acc)._reindex()


def merge(peaks_a, *peaks_b, **kwargs):
    """Combine two :class:`DeconvolutedPeakSet` objects.

    Parameters
    ----------
    peaks_a: :class:`DeconvolutedPeakSet`
    peaks_b: :class:`DeconvolutedPeakSet`
    copy: bool
        Whether or not to copy the peaks first. If not,
        the two input peak sets should not be used again
        as their indices will have been corrupted.

    Returns
    -------
    :class:`DeconvolutedPeakSet`
    """
    copy = kwargs.get("copy", True)
    tp = peaks_a.__class__
    if copy:
        peaks = ()
        peaks += tuple(p.clone() for p in peaks_a)
        for b in peaks_b:
            peaks += tuple(p.clone() for p in b)
    else:
        peaks = tuple(peaks_a)
        for b in peaks_b:
            peaks += tuple(b)
    peaks = tp(peaks)
    try:
        peaks.reindex()
    except AttributeError:
        pass
    return peaks


mz_getter = operator.attrgetter('mz')
neutral_mass_getter = operator.attrgetter("neutral_mass")


def _get_nearest_peak(peaklist, neutral_mass, use_mz=False):
    lo = 0
    hi = len(peaklist)

    tol = 1

    getter = operator.attrgetter("mz") if use_mz else operator.attrgetter("neutral_mass")

    def sweep(lo, hi):
        best_error = float('inf')
        best_index = None
        for i in range(hi - lo):
            i += lo
            v = getter(peaklist[i])
            err = abs(v - neutral_mass)
            if err < best_error:
                best_error = err
                best_index = i
        return peaklist[best_index], best_error

    def binsearch(lo, hi):
        if (hi - lo) < 5:
            return sweep(lo, hi)
        else:
            mid = (hi + lo) // 2
            v = getter(peaklist[mid])
            if abs(v - neutral_mass) < tol:
                return sweep(lo, hi)
            elif v > neutral_mass:
                return binsearch(lo, mid)
            else:
                return binsearch(mid, hi)
    return binsearch(lo, hi)


def _sweep_solution(array, value, mid, tolerance, getter=operator.attrgetter('neutral_mass')):

    best_index = mid
    best_error = float('inf')
    n = len(array)

    i = 0
    while mid - i >= 0 and i <= mid:
        target = array[mid - i]
        abs_error = abs(ppm_error(value, getter(target)))
        if abs_error < tolerance:
            if abs_error < best_error:
                best_index = mid - i
                best_error = abs_error
        else:
            break
        i += 1
    i = 1
    while (mid + i) < (n - 1):
        target = array[mid + i]
        abs_error = abs(ppm_error(value, getter(target)))
        if abs_error < tolerance:
            if abs_error < best_error:
                best_index = mid + i
                best_error = abs_error
        else:
            break
        i += 1
    if best_error == float('inf'):
        return None
    else:
        return array[best_index]


def binary_search(peak_set, neutral_mass, tolerance, getter=operator.attrgetter('neutral_mass')):

    lo = 0
    hi = len(peak_set)

    while hi != lo:
        mid = (hi + lo) // 2
        found_peak = peak_set[mid]
        found_mass = getter(found_peak)

        if abs(ppm_error(found_mass, neutral_mass)) < tolerance:
            found_peak = _sweep_solution(peak_set, neutral_mass, mid, tolerance)
            return found_peak
        elif hi - lo == 1:
            return None
        elif found_mass > neutral_mass:
            hi = mid
        else:
            lo = mid


try:
    has_c = True
    _Envelope = Envelope
    __Index = _Index
    _DeconvolutedPeak = DeconvolutedPeak
    _DeconvolutedPeakSolution = DeconvolutedPeakSolution
    _DeconvolutedPeakSet = DeconvolutedPeakSet

    from ms_deisotope._c.peak_set import (
        Envelope, _Index, DeconvolutedPeak, DeconvolutedPeakSolution,
        DeconvolutedPeakSetIndexed as DeconvolutedPeakSet
    )

    from ms_deisotope._c.peak_set import IonMobilityDeconvolutedPeak

except ImportError:
    has_c = False


def window_peak_set(peak_set: DeconvolutedPeakSet, window_size: float=100.0, peaks_per_window: int=10) -> List[List[DeconvolutedPeak]]:
    """Generate non-overlapping windows over a :class:`~.DeconvolutedPeakSet`, selecting the most
    abundant peaks in each window.

    Parameters
    ----------
    peak_set : DeconvolutedPeakSet
        The peak set to generate windows from.
    window_size : float
        The width of the window in daltons. The default is 100.
    peaks_per_window : int
        The number of peaks to retain per window, by intensity. Default is 10.

    Returns
    -------
    List[List[DeconvolutedPeak]]
    """
    lower_bound = math.floor(peak_set[0].neutral_mass / window_size) * window_size
    upper_bound = math.ceil(peak_set[-1].neutral_mass / window_size) * window_size

    window_count = int(math.ceil((upper_bound - lower_bound) / window_size))
    windows = []

    window_upper_bound = lower_bound + window_size
    window_lower_bound = lower_bound

    for _ in range(window_count):
        window = peak_set.between(window_lower_bound, window_upper_bound)
        window = sorted(
            sorted(window, key=lambda x: x.intensity, reverse=True)[:peaks_per_window],
            key=lambda x: x.neutral_mass)
        windows.append(window)
        window_lower_bound += window_size
        window_upper_bound += window_size
    return windows


def decharge(self, include_envelopes=True, charge_carrier=PROTON, new_charge=1):
    """Transfomr multiply charged deconvoluted peaks into singly charged deconvoluted peaks.

    This operation returns a copy of the peak set, the input peak set is unchanged.

    Parameters
    ----------
    include_envelopes : bool, optional
        Whether or not to charge-transform the isotopic envelope fitted in :attr:`DeconvolutedPeak.envelope`
        (the default is True)
    charge_carrier : float, optional
        The charge carrier mass to use when recalculating m/z (the default is PROTON)
    new_charge : int, optional
        The new charge to set all peaks to, the default being +1


    Returns
    -------
    DeconvolutedPeakSet
    """
    peaks = []
    for peak in self:
        mz = mass_charge_ratio(peak.neutral_mass, new_charge,
                               charge_carrier=charge_carrier)
        if include_envelopes:
            envelope = Envelope(
                [EnvelopePair(
                    mass_charge_ratio(
                        calc_neutral_mass(p.mz, peak.charge, charge_carrier),
                        new_charge, charge_carrier), p.intensity)
                 for p in peak.envelope])
        else:
            envelope = peak.envelope
        new_peak = DeconvolutedPeak(
            peak.neutral_mass, peak.intensity, new_charge, peak.signal_to_noise, None,
            peak.full_width_at_half_max, peak.a_to_a2_ratio, peak.most_abundant_mass,
            peak.average_mass, peak.score, envelope, mz, peak.fit, peak.chosen_for_msms,
            peak.area)
        peaks.append(new_peak)
    new_peak_set = self.__class__(peaks)
    new_peak_set.reindex()
    return new_peak_set


def envelopes_to_peak_set(self):
    """Convert a set of deconvoluted peaks with fitted isotopic envelopes into a
    set of centroids representing those envelope peaks.

    Returns
    -------
    :class:`ms_peak_picker.PeakSet`
    """
    peaks = []
    for peak in self:
        for point in peak.envelope:
            peaks.append(
                simple_peak(point.mz, point.intensity, peak.full_width_at_half_max))
    new_peak_set = FittedPeakSet(peaks)
    new_peak_set.reindex()
    return new_peak_set


class IonMobilityDeconvolutedPeakSet(DeconvolutedPeakSet):
    def reindex(self):
        """
        Updates the :attr:`index` of each peak in `self` and updates the
        sorted order.

        Returns
        -------
        self: DeconvolutedPeakSet
        """
        self.peaks = tuple(
            sorted(self.peaks, key=lambda x: (x.neutral_mass, x.drift_time)))
        self._mz_ordered = tuple(
            sorted(self.peaks, key=lambda x: (x.mz, x.drift_time)))
        for i, peak in enumerate(self.peaks):
            peak.index = _Index()
            peak.index.neutral_mass = i
        for i, peak in enumerate(self._mz_ordered):
            peak.index.mz = i
        return self


if TYPE_CHECKING:
    from ms_deisotope.feature_map.feature_fit import IonMobilityProfileDeconvolutedLCMSFeature


class IonMobilityProfileDeconvolutedPeakSolution(DeconvolutedPeakSolution):
    ion_mobility_interval: SimpleInterval

    def __init__(self, solution, fit=None, *args, **kwargs):
        super().__init__(solution, fit=fit, *args, **kwargs)
        self.ion_mobility_interval = SimpleInterval(
            solution.start_time, solution.end_time)

    def overlaps(self, other: 'IonMobilityProfileDeconvolutedPeakSolution') -> bool:
        return self.ion_mobility_interval.overlaps(other.ion_mobility_interval)

    @classmethod
    def from_feature(cls, feature: 'IonMobilityProfileDeconvolutedLCMSFeature', include_mz: bool=False, include_envelope: bool=True) -> 'IonMobilityProfileDeconvolutedPeakSolution':
        return cls(feature, fit=None, neutral_mass=feature.neutral_mass,
                   intensity=feature.intensity, charge=feature.charge,
                   envelope=feature.sum_envelopes() if include_envelope else None,
                   signal_to_noise=feature.intensity, index=0, full_width_at_half_max=0,
                   a_to_a2_ratio=0, most_abundant_mass=0, average_mass=0,
                   score=feature.score, chosen_for_msms=False,
                   mz=feature.mz if include_mz else 0.0)



from ms_deisotope._c.peak_dependency_network.intervals cimport SpanningMixin, IntervalTreeNode

from ms_deisotope._c.feature_map.lcms_feature cimport FeatureBase, EmptyFeature
from ms_deisotope._c.feature_map.feature_fit cimport LCMSFeatureSetFit, map_coord


cdef class SpanningMixin2D(object):
    cdef:
        public map_coord start
        public map_coord end

    cdef bint _contains(self, map_coord i)

    cpdef bint contains(self, map_coord i)

    cpdef bint overlaps(self, SpanningMixin2D interval)

    cpdef bint is_contained_in_interval(self, SpanningMixin2D interval)

    cpdef bint contains_interval(self, SpanningMixin2D interval)


cdef class FeatureNode(object):
    cdef:
        public FeatureBase feature
        public dict links
        Py_hash_t _hash


cdef class DependenceCluster(SpanningMixin2D):
    cdef:
        public object parent
        public list dependencies
        public LCMSFeatureSetFit best_fit
        public bint maximize

    cpdef _reset(self)
    cpdef add(self, LCMSFeatureSetFit fit)
    cpdef disjoint_subset(self, int max_size=*)
    cpdef list disjoint_best_fits(self, int max_size=*)

    cdef LCMSFeatureSetFit _best_fit(self)
    cdef map_coord _start(self)
    cdef map_coord _end(self)


cdef class FeatureSetFitNode(SpanningMixin2D):
    cdef:
        public LCMSFeatureSetFit fit
        public set edges
        public set overlap_edges
        public set feature_indices
        public Py_hash_t _hash
        public long index
        public double score

    cdef void _init_indices(self)

    cpdef visit(self, FeatureSetFitNode other)
    cpdef bint isdisjoint(self, FeatureSetFitNode other)


cdef class ConnectedSubgraph(object):
    cdef:
        public tuple nodes
        public bint maximize

    cpdef populate_edges(self)
    cpdef find_heaviest_path(self, method=*)
    cdef void _init(self)


cdef class FeatureDependenceGraphBase(object):
    cdef:
        public dict nodes
        public object feature_map
        public set dependencies
        public list clusters

        public int max_missing_features
        public bint use_monoisotopic_superceded_filtering
        public bint maximize

    cpdef list nodes_for(self, LCMSFeatureSetFit fit_record, dict cache=*)
    cpdef add_fit_dependence(self, LCMSFeatureSetFit fit_record)
    cpdef drop_superceded_fits(self)
    cpdef drop_fit_dependence(self, LCMSFeatureSetFit fit_record, dict cache=*)
    cpdef best_exact_fits(self)
    cpdef drop_gapped_fits(self, n=*)
    cpdef _gather_independent_clusters(self, dict nodes_for_cache=*)





"""Defines types for describing mass spectrometry data transformations,
data processing sequences, and a database of controlled vocabulary terms
for them.

.. note::
    The entities in this module are only for *naming* these transformations
    and how they were combined to process mass spectra. For implementations,
    see other features of :mod:`ms_deisotope` and :mod:`ms_peak_picker`, or
    other libraries.
"""

from typing import Any, List, Union, OrderedDict

from six import string_types as basestring

from ms_deisotope.utils import uid

from .cv import Term, TermSet
# from .software import Software, SoftwareName

DataTransformationOrStr = Union[str, 'DataTransformation']


class DataTransformation(Term):
    """Describes a named data transformation, either
    using a controlled-vocabulary term or user-defined name.

    A :class:`DataTransformation` is equal to its name and its controlled
    vocabulary identifier.
    """
    pass


data_transformations = []


# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('data transformation', term_cls_name="DataTransformation", writer=cog.out)
# ]]]
# CV Version: 4.1.95
data_transformations = TermSet([
    DataTransformation('file format conversion', 'MS:1000530',
                       ('Conversion of one file format to another.'),
                       'data transformation',
                       ['data transformation']),
    DataTransformation('data processing action', 'MS:1000543',
                       ('Data processing attribute used to describe the type of data '
                        'processing performed on the data file.'),
                       'data transformation',
                       ['data transformation']),
    DataTransformation('Conversion to mzML', 'MS:1000544',
                       ('Conversion of a file format to Proteomics Standards '
                        'Initiative mzML file format.'),
                       'data transformation',
                       ['file format conversion', 'data transformation']),
    DataTransformation('Conversion to mzXML', 'MS:1000545',
                       ('Conversion of a file format to Institute of Systems Biology '
                        'mzXML file format.'),
                       'data transformation',
                       ['file format conversion', 'data transformation']),
    DataTransformation('Conversion to mzData', 'MS:1000546',
                       ('Conversion of a file format to Proteomics Standards '
                        'Initiative mzData file format.'),
                       'data transformation',
                       ['file format conversion', 'data transformation']),
    DataTransformation('Conversion to dta', 'MS:1000741',
                       ('Conversion to dta format.'),
                       'data transformation',
                       ['file format conversion', 'data transformation']),
    DataTransformation('Conversion to mzMLb', 'MS:1002839',
                       ('Conversion of a file format to Proteomics Standards '
                        'Initiative mzMLb file format.'),
                       'data transformation',
                       ['file format conversion', 'data transformation']),
    DataTransformation('deisotoping', 'MS:1000033',
                       ('The removal of isotope peaks to represent the fragment ion '
                        'as one data point and is commonly done to reduce complexity. '
                        'It is done in conjunction with the charge state '
                        'deconvolution.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('charge deconvolution', 'MS:1000034',
                       ('The determination of the mass of an ion based on the mass '
                        'spectral peaks that represent multiple-charge ions.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('peak picking', 'MS:1000035',
                       ('Spectral peak processing conducted on the acquired data to '
                        'convert profile data to centroided data.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('smoothing', 'MS:1000592',
                       ('A process of reducing spikes of intensity in order to reduce '
                        'noise while preserving real peak signal. Many algorithms can '
                        'be applied for this process.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('baseline reduction', 'MS:1000593',
                       ('A process of removal of varying intensities generated due to '
                        'variable energy absorption before further processing can '
                        'take place. Baseline reduction facilitates meaningful '
                        'comparision between intensities of m/z values.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('retention time alignment', 'MS:1000745',
                       ('The correction of the spectrum scan times, as used e.g. in '
                        'label-free proteomics.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('charge state calculation', 'MS:1000778',
                       ("A process that infers the charge state of an MSn spectrum's "
                        'precursor(s) by the application of some algorithm.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('precursor recalculation', 'MS:1000780',
                       ('A process that recalculates existing precursor selected ions '
                        'with one or more algorithmically determined precursor '
                        'selected ions.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('intensity normalization', 'MS:1001484',
                       ('Normalization of data point intensities.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('m/z calibration', 'MS:1001485',
                       ('Calibration of data point m/z positions.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('data filtering', 'MS:1001486',
                       ('Filtering out part of the data.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('adduct deconvolution', 'MS:1003220',
                       ('Data processing action of merging of the measurements of '
                        'potentially multiple adducts into a single representation '
                        'that is independent of the small ion that adds charge to a '
                        'larger molecule.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('ion mobility deconvolution', 'MS:1003222',
                       ('Data processing action of merging multiple ion peaks '
                        'acquired at different ion mobility steps into a single mass '
                        'spectrum representing a single analyte.'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('peak intensity transform', 'MS:1003240',
                       ('A mathematical transformation applied to peak intensities, '
                        'for example, as a way to modify the weight put on each peak '
                        'when computing spectral match scores'),
                       'data transformation',
                       ['data processing action', 'data transformation']),
    DataTransformation('area peak picking', 'MS:1000801',
                       ('Spectral peak processing conducted on the acquired data to '
                        'convert profile data to centroided data. The area defined by '
                        'all raw data points that belong to the peak is reported.'),
                       'data transformation',
                       ['peak picking', 'data processing action', 'data transformation']),
    DataTransformation('height peak picking', 'MS:1000802',
                       ('Spectral peak processing conducted on the acquired data to '
                        'convert profile data to centroided data. The maximum '
                        'intensity of all raw data points that belong to the peak is '
                        'reported.'),
                       'data transformation',
                       ['peak picking', 'data processing action', 'data transformation']),
    DataTransformation('Savitzky-Golay smoothing', 'MS:1000782',
                       ('Reduces intensity spikes by applying local polynomial '
                        'regression (of degree k) on a distribution (of at least k+1 '
                        'equally spaced points) to determine the smoothed value for '
                        'each point. It tends to preserve features of the '
                        'distribution such as relative maxima, minima and width, '
                        "which are usually 'flattened' by other adjacent averaging "
                        'techniques.'),
                       'data transformation',
                       ['smoothing', 'data processing action', 'data transformation']),
    DataTransformation('LOWESS smoothing', 'MS:1000783',
                       ('Reduces intensity spikes by applying a modelling method '
                        'known as locally weighted polynomial regression. At each '
                        'point in the data set a low-degree polynomial is fit to a '
                        'subset of the data, with explanatory variable values near '
                        'the point whose response is being estimated. The polynomial '
                        'is fit using weighted least squares, giving more weight to '
                        'points near the point whose response is being estimated and '
                        'less weight to points further away. The value of the '
                        'regression function for the point is then obtained by '
                        'evaluating the local polynomial using the explanatory '
                        'variable values for that data point. The LOESS fit is '
                        'complete after regression function values have been computed '
                        'for each of the n data points. Many of the details of this '
                        'method, such as the degree of the polynomial model and the '
                        'weights, are flexible.'),
                       'data transformation',
                       ['smoothing', 'data processing action', 'data transformation']),
    DataTransformation('Gaussian smoothing', 'MS:1000784',
                       ('Reduces intensity spikes by convolving the data with a one- '
                        'dimensional Gaussian function.'),
                       'data transformation',
                       ['smoothing', 'data processing action', 'data transformation']),
    DataTransformation('moving average smoothing', 'MS:1000785',
                       ('Reduces intensity spikes by averaging each point with two or '
                        'more adjacent points. The more adjacent points that used, '
                        'the stronger the smoothing effect.'),
                       'data transformation',
                       ['smoothing', 'data processing action', 'data transformation']),
    DataTransformation('wavelet transformation smoothing', 'MS:1001997',
                       ('The random noise is removed by using the undecimated wavelet '
                        'transform." [DOI:10.1093/bioinformatics/btl355'),
                       'data transformation',
                       ['smoothing', 'data processing action', 'data transformation']),
    DataTransformation('top hat baseline reduction', 'MS:1001994',
                       ('Top-hat morphological filter based on the basic '
                        "morphological operations 'erosion' and 'dilatation'."),
                       'data transformation',
                       ['baseline reduction', 'data processing action', 'data transformation']),
    DataTransformation('convex hull baseline reduction', 'MS:1001995',
                       ('Constructs the baseline by fitting multiple parabolas to the '
                        'spectrum starting with the large scale structures.'),
                       'data transformation',
                       ['baseline reduction', 'data processing action', 'data transformation']),
    DataTransformation('median baseline reduction', 'MS:1001996',
                       ('The spectrum that will be baseline subtracted is divided '
                        'into a number of segments.'),
                       'data transformation',
                       ['baseline reduction', 'data processing action', 'data transformation']),
    DataTransformation('below precursor intensity dominance charge state calculation', 'MS:1000779',
                       ('Infers charge state as single or ambiguously multiple by '
                        'determining the fraction of intensity below the precursor '
                        'm/z.'),
                       'data transformation',
                       ['charge state calculation', 'data processing action', 'data transformation']),
    DataTransformation('msPrefix precursor recalculation', 'MS:1000781',
                       ('Recalculates one or more precursor selected ions by peak '
                        'detection in the isolation windows of high accuracy MS '
                        'precursor scans.'),
                       'data transformation',
                       ['precursor recalculation', 'data processing action', 'data transformation']),
    DataTransformation('area normalization', 'MS:1001999',
                       ('Normalization of areas below the curves.'),
                       'data transformation',
                       ['intensity normalization', 'data processing action', 'data transformation']),
    DataTransformation('low intensity data point removal', 'MS:1000594',
                       ('The removal of very low intensity data points that are '
                        'likely to be spurious noise rather than real signal.'),
                       'data transformation',
                       ['data filtering', 'data processing action', 'data transformation']),
    DataTransformation('high intensity data point removal', 'MS:1000746',
                       ('The removal of very high intensity data points.'),
                       'data transformation',
                       ['data filtering', 'data processing action', 'data transformation']),
    DataTransformation('square root transform', 'MS:1003241',
                       ('A mathematical transformation applied to peak intensities, '
                        'in which peak intensities are replaced by their square roots'),
                       'data transformation',
                       ['peak intensity transform', 'data processing action', 'data transformation']),
    DataTransformation('rank transform', 'MS:1003242',
                       ('A mathematical transformation applied to peak intensities, '
                        'in which peak intensities are replaced by their ranks'),
                       'data transformation',
                       ['peak intensity transform', 'data processing action', 'data transformation']),
    DataTransformation('sophisticated numerical annotation procedure', 'MS:1001998',
                       ('It searches for known patterns in the measured spectrum." '
                        '[DOI:10.1021/ac951158i'),
                       'data transformation',
                       ['area peak picking', 'peak picking', 'data processing action', 'data transformation']),
])
# [[[end]]]


def data_transformation(name):
    """Translate a given name or identifier into a :class:`DataTransformation`
    instance.

    If no match is found in the database of known :class:`DataTransformation`
    types, a new dummy :class:`DataTransformation` is returned with all fields
    set to the value of ``name``

    Returns
    -------
    DataTransformation
    """
    try:
        return data_transformations[name]
    except KeyError:
        return DataTransformation(name, name, name, name, [name])


class ProcessingMethod(object):
    """Describes a single action applied to the associated spectra, composed of one
    or more data transformation operations.

    Attributes
    ----------
    operations: :class:`~.OrderedDict`
        The data transformations applied, mapping transformation term to any
        parameter value or the empty string.
    software_id: str
        A reference to a particular version of software used to carry out the
        transformations.
    order: int
        The order in which this action was performed, in the context of the
        parent :class:`DataProcessingInformation` object

    See Also
    --------
    :class:`DataTransformation`
    """
    operations: OrderedDict[
        DataTransformation,
        Union[Any, List[Any]]
    ]
    software_id: str
    order: int

    def __init__(self, operations=None, software_id=None, order=0):
        self.operations = operations or OrderedDict()
        self.software_id = software_id
        self.order = order

    def __getitem__(self, i: Union[int, slice]):
        return list(self.operations.items())[i]

    def __len__(self):
        return len(self.operations)

    def __iter__(self):
        return iter(self.operations.items())

    def _get_names(self):
        return ["%r: %r" % (t.name, v) for t, v in self.operations.items()]

    def __repr__(self):
        template = "{self.__class__.__name__}({names}, order={self.order})"
        return template.format(self=self, names='[%s]' % ', '.join(self._get_names()))

    def remove(self, operation: DataTransformationOrStr):
        return self.operations.pop(operation, None)

    def add(self, operation: DataTransformationOrStr, value: Any = ''):
        """Add a new :class:`DataTransformation` operation to this method.

        Parameters
        ----------
        operation: :class:`DataTransformation` or :class:`str`
            The transformation applied. If passed as a :class:`str`, it will be translated
            into a :class:`DataTransformation`.
        value: :class:`object`
            A parameter value associated with the operation. Defaults to the empty string.
        """
        if isinstance(operation, basestring):
            operation = data_transformation(operation)
        if operation in self.operations:
            val = self.operations[operation]
            if not isinstance(val, list):
                self.operations[operation] = [val, value]
            else:
                val.append(value)
        else:
            self.operations[operation] = value

    def get(self, operation: DataTransformationOrStr, default: Any = None) -> Any:
        try:
            return self.operations[operation]
        except KeyError:
            for k, v in self:
                if k == operation:
                    return v
            else:
                return default

    def update(self, operations):
        """Add each operation in ``operations`` to this method.

        Parameters
        ----------
        operations: :class:`~.Mapping` or :class:`Iterable`
            If given a :class:`~.Mapping`, add each key-value
            pair using :meth:`add`. Otherwise extract keys or key-value pairs
            if :class:`Iterable` is supported, adding each in succession
        """
        if isinstance(operations, dict):
            for op, val in operations.items():
                self.add(op, val)
        else:
            for op in operations:
                if isinstance(op, dict):
                    self.add(op['name'], op['value'])
                elif isinstance(op, (tuple, list)) and len(op) == 2:
                    self.add(*op)
                else:
                    self.add(op)


class DataProcessingInformation(object):
    """A :class:`Sequence` of :class:`ProcessingMethod` instances, describing the sequence
    of data transformations applied to the spectra in the data associated with this instance.

    Attributes
    ----------
    methods: list
        A list of :class:`ProcessingMethod` instances describing the individual operations done
        in an ordered fashion.
    id: int
        A within-source unique identifier
    """

    def __init__(self, methods=None, id=None):
        self.methods = methods or []
        self.id = id or uid()

    def __getitem__(self, i):
        return self.methods[i]

    def __len__(self):
        return len(self.methods)

    def __iter__(self):
        return iter(self.methods)

    def __repr__(self):
        template = "{self.__class__.__name__}({self.methods}, id={self.id!r})"
        return template.format(self=self)


__all__ = [
    "DataTransformation", "data_transformations", 'data_transformation',
    "ProcessingMethod", "DataProcessingInformation"
]



"""
Writing mzMLb
-------------

Using the :mod:`psims` library, :mod:`ms_deisotope.output.mzmlb` can write an mzMLb
file with all associated metadata. This uses the same machinery as :mod:`ms_deisotope.output.mzml`
to accomplish this.

This module also contains a specialized version of :class:`~.MzMLbLoader`,
:class:`~.ProcessedMzMLbLoader`, which can directly reconstruct each
deconvoluted peak list and provides fast access to an extended index of
metadata that :class:`~.MzMLbSerializer` writes to an external file.

All behavior available for the :class:`~.ProcessedMzMLLoader` are available for
:class:`~.ProcessedMzMLbLoader`.


.. code:: python

    import ms_deisotope
    from ms_deisotope.test.common import datafile
    from ms_deisotope.output.mzmlb import MzMLbSerializer

    reader = ms_deisotope.MSFileLoader(datafile("small.mzML"))
    with MzMLbSerializer('small.deconvoluted.mzMLb', n_spectra=len(reader)) as writer:
        writer.copy_metadata_from(reader)
        for bunch in reader:
            bunch.precursor.pick_peaks()
            bunch.precursor.deconvolute()
            for product in bunch.products:
                product.pick_peaks()
                product.deconvolute()
            writer.save(bunch)

"""

import numpy as np

from psims.mzml.writer import COMPRESSION_ZLIB

try:
    from psims.mzmlb.writer import MzMLbWriter as _MzMLbWriter
    is_available = True
except ImportError:
    _MzMLbWriter = None
    is_available = False

from ms_deisotope.data_source.mzmlb import MzMLbLoader

from .common import LCMSMSQueryInterfaceMixin, ScanDeserializerBase, SampleRun
from .mzml import MzMLSerializer as _MzMLSerializer, PeakSetDeserializingMixin, _IonMobility3DSerializerBase


class MzMLbSerializer(_MzMLSerializer):

    file_extensions = {
        "mzmlb",
    }

    _format_conversion_term = "Conversion to mzMLb"

    def _make_writer(self, handle):
        compression = self.compression
        compression_opts = None
        if self.compression == COMPRESSION_ZLIB:
            compression = 'gzip'
            compression_opts = 4
        self.compression = 'none'
        return _MzMLbWriter(
            self.handle,
            close=self._should_close,
            h5_compression=compression,
            h5_compression_options=compression_opts
        )


class IonMobilityAware3DMzMLbSerializer(_IonMobility3DSerializerBase, MzMLbSerializer):
    default_data_encoding = MzMLbSerializer.default_data_encoding.copy()
    default_data_encoding.update({
        "feature id array": np.int32,
    })


class ProcessedMzMLbLoader(PeakSetDeserializingMixin, MzMLbLoader, ScanDeserializerBase, LCMSMSQueryInterfaceMixin):

    file_extensions = {
        "mzmlb",
    }

    def __init__(self, source_file, use_index=True, use_extended_index=True):
        super(ProcessedMzMLbLoader, self).__init__(
            source_file, use_index=use_index, decode_binary=True)
        self.extended_index = None
        self._scan_id_to_rt = dict()
        self._sample_run = None
        self._use_extended_index = use_extended_index
        if self._use_index:
            if self._use_extended_index:
                self.require_extended_index()

    def _dispose(self):
        self._scan_id_to_rt.clear()
        self.extended_index.clear()
        super(ProcessedMzMLbLoader, self)._dispose()

    def __reduce__(self):
        return self.__class__, (self.source_file, self._use_index, self._use_extended_index)

    def _make_sample_run(self):
        samples = self.samples()
        sample = samples[0]
        return SampleRun(name=sample.name, uuid=sample['SampleRun-UUID'], **dict(sample.items()))

    @property
    def sample_run(self):
        if self._sample_run is None:
            self._sample_run = self._make_sample_run()
        return self._sample_run


ProcessedMzMLbDeserializer = ProcessedMzMLbLoader
